#!/usr/bin/python
# -*- coding: utf-8 -*-

import datetime
import typing
import unittest

from sane_finances.communication.downloader import DownloadStringResult
from sane_finances.sources.base import DownloadParameterValuesStorage
from sane_finances.sources.base import (
    InstrumentStringDataDownloader,
    InstrumentHistoryDownloadParameters)
from .fakes import FakeInstrumentHistoryDownloadParameters


class SomeDownloadParameterValuesStorage(DownloadParameterValuesStorage):

    def is_dynamic_enum_type(self, cls: type) -> bool:
        return False

    def get_all_managed_types(self) -> typing.Iterable[typing.Type]:
        return ()

    def get_dynamic_enum_key(self, instance):
        return None


class SomeInstrumentStringDataDownloader(InstrumentStringDataDownloader):

    def download_instrument_history_string(
            self,
            parameters: InstrumentHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime) -> DownloadStringResult:
        return DownloadStringResult('')

    def download_instruments_info_string(self, parameters) -> DownloadStringResult:
        return DownloadStringResult('')


class TestDownloadParameterValuesStorage(unittest.TestCase):
    """ Test default implementation of DownloadParameterValuesStorage base class
    """

    def test_reload_Success(self):
        storage = SomeDownloadParameterValuesStorage()

        storage.reload()

    def test_get_dynamic_enum_value_by_key_AlwaysReturnNone(self):
        storage = SomeDownloadParameterValuesStorage()

        self.assertIsNone(storage.get_dynamic_enum_value_by_key(int, 1))
        self.assertIsNone(storage.get_dynamic_enum_value_by_key(str, ''))
        self.assertIsNone(storage.get_dynamic_enum_value_by_key(DownloadParameterValuesStorage, None))

    def test_get_dynamic_enum_value_by_choice_AlwaysReturnNone(self):
        storage = SomeDownloadParameterValuesStorage()

        self.assertIsNone(storage.get_dynamic_enum_value_by_choice(int, 1))
        self.assertIsNone(storage.get_dynamic_enum_value_by_choice(str, ''))
        self.assertIsNone(storage.get_dynamic_enum_value_by_choice(DownloadParameterValuesStorage, None))

    def test_get_all_parameter_values_for_AlwaysReturnNone(self):
        storage = SomeDownloadParameterValuesStorage()

        self.assertIsNone(storage.get_all_parameter_values_for(int))
        self.assertIsNone(storage.get_all_parameter_values_for(str))
        self.assertIsNone(storage.get_all_parameter_values_for(DownloadParameterValuesStorage))

    def test_get_parameter_type_choices_AlwaysReturnNone(self):
        storage = SomeDownloadParameterValuesStorage()

        self.assertIsNone(storage.get_parameter_type_choices(int))
        self.assertIsNone(storage.get_parameter_type_choices(str))
        self.assertIsNone(storage.get_parameter_type_choices(DownloadParameterValuesStorage))


class TestInstrumentStringDataDownloader(unittest.TestCase):
    """ Test default implementation of InstrumentStringDataDownloader base class
    """

    def test_adjust_date_from_to_holidays_Success(self):
        storage = SomeInstrumentStringDataDownloader()

        date_from = datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc)  # New Year's Eve
        adjusted_date_from = storage.adjust_date_from_to_holidays(date_from)

        self.assertEqual(adjusted_date_from.year, date_from.year - 1)

        date_from = datetime.datetime(2000, 12, 25, tzinfo=datetime.timezone.utc)  # Christmas holidays
        adjusted_date_from = storage.adjust_date_from_to_holidays(date_from)

        self.assertLess(adjusted_date_from.day, date_from.day)

    def test_adjust_download_instrument_history_parameters_RaiseWhenWrongDates(self):
        storage = SomeInstrumentStringDataDownloader()

        params = FakeInstrumentHistoryDownloadParameters()
        moment_from = datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc)
        moment_to = moment_from - datetime.timedelta(days=1)

        self.assertGreater(moment_from, moment_to)

        with self.assertRaises(ValueError):
            _ = storage.adjust_download_instrument_history_parameters(
                params,
                moment_from,
                moment_to)

    def test_paginate_download_instrument_history_parameters_DoNothing(self):
        storage = SomeInstrumentStringDataDownloader()

        params = FakeInstrumentHistoryDownloadParameters()
        moment_from = datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc)
        moment_to = moment_from + datetime.timedelta(days=1)

        paginated_params = list(storage.paginate_download_instrument_history_parameters(
            params,
            moment_from,
            moment_to))

        self.assertSequenceEqual(paginated_params, [(params, moment_from, moment_to)])



""" Sources from Moscow Exchange.
"""
from . import v1_3
from ..generic import InstrumentExporterRegistry, register_instrument_history_values_exporter

__all__ = ['v1_3']

register_instrument_history_values_exporter(
    InstrumentExporterRegistry(
        name=v1_3.exporters.MoexIndexExporterFactory_v1_3.name,
        provider_site=v1_3.exporters.MoexIndexExporterFactory_v1_3.provider_site,
        api_url=v1_3.exporters.MoexIndexExporterFactory_v1_3.api_url,
        factory=v1_3.exporters.MoexIndexExporterFactory_v1_3()
    ))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Metadata for export index data from app2.msci.com
"""

import decimal
import typing
import datetime
import enum
import dataclasses

from ...base import (
    InstrumentValue, InstrumentInfo, InstrumentValueProvider, InstrumentInfoProvider,
    InstrumentHistoryDownloadParameters, DownloadParametersFactory)
from ...inspection import InstrumentInfoParameter
from ....annotations import Volatile, LEGACY_ANNOTATIONS, Description

if LEGACY_ANNOTATIONS:  # pragma: no cover
    from ....annotations import Annotated
else:  # pragma: no cover
    from typing import Annotated


class Scopes(enum.Enum):
    """ Index scope.
    """

    def __new__(cls, value: str, description: str):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = str(description)
        return obj

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}: " \
               f"'{self.value}' ('{self.description}')>"

    REGIONAL = ('R', 'Regional')
    COUNTRY = ('C', 'Country')


class Markets(enum.Enum):
    """ Market of index.
    """

    def __new__(cls, value: str, description: str, scope: Scopes):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = str(description)

        # see https://github.com/PyCQA/pylint/issues/1801 for pylint disable hint details
        obj.scope = Scopes(scope)  # pylint: disable=no-value-for-parameter
        return obj

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}: " \
               f"'{self.value}' ('{self.description}', '{self.scope.description}')>"

    REGIONAL_ALL_COUNTRY = ('1896', 'All Country (DM+EM)', Scopes.REGIONAL)
    REGIONAL_CHINA_MARKETS = ('2809', 'China Markets', Scopes.REGIONAL)
    REGIONAL_DEVELOPED_MARKETS = ('1897', 'Developed Markets (DM)', Scopes.REGIONAL)
    REGIONAL_EMERGING_MARKETS = ('1898', 'Emerging Markets (EM)', Scopes.REGIONAL)
    REGIONAL_FRONTIER_MARKETS = ('2115', 'Frontier Markets (FM)', Scopes.REGIONAL)
    REGIONAL_GCC_AND_ARABIAN_MARKETS = ('1899', 'GCC and Arabian Markets', Scopes.REGIONAL)

    COUNTRY_CHINA_MARKETS = ('2810', 'China Markets', Scopes.COUNTRY)
    COUNTRY_DEVELOPED_MARKETS = ('1900', 'Developed Markets (DM)', Scopes.COUNTRY)
    COUNTRY_EMERGING_MARKETS = ('1901', 'Emerging Markets (EM)', Scopes.COUNTRY)
    COUNTRY_FRONTIER_MARKETS = ('2114', 'Frontier Markets (FM)', Scopes.COUNTRY)
    COUNTRY_GCC_AND_ARABIAN_MARKETS = ('1902', 'GCC and Arabian Markets', Scopes.COUNTRY)


class Formats(enum.Enum):
    """ Response format.
    """
    XML = 'XML'
    CSV = 'CSV'

    @staticmethod
    def get_file_extension(response_format: 'Formats') -> str:
        """ Get suggested file extension (with dot).

        :param response_format: Response format
        :return: Suggested file extension (with dot).
        """
        return {Formats.XML: '.xml', Formats.CSV: '.csv'}[response_format]


class Currencies(enum.Enum):
    """ Index currency.

    The MSCI Country and Regional Indices are calculated in local currency as well as in USD.
    The concept of a “local currency” calculation excludes the impact of currency fluctuations.
    """

    def __new__(cls, value: str, description: str):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = str(description)
        return obj

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}: " \
               f"'{self.value}' ('{self.description}')>"

    LOCAL = ('0', 'Local')
    USD = ('15', 'United States Dollar')
    EUR = ('119', 'Euro')
    GBP = ('18', 'British Pound Sterling')
    JPY = ('10', 'Japanese Yen')
    CAD = ('16', 'Canadian Dollar')
    CHF = ('3', 'Swiss Franc')
    HKD = ('11', 'Hong Kong Dollar')
    AUD = ('1', 'Australian Dollar')


class IndexLevels(enum.Enum):
    """ Index level (or price level).

    Total return indices measure the market performance,
    including price performance and income from regular cash distributions
    (cash dividend payments or capital repayments).

    Gross Daily Total Return: This series approximates the maximum possible reinvestment
    of regular cash distributions (dividends or capital repayments).
    The amount reinvested is the cash distributed to individuals resident in the country of the company,
    but does not include tax credits.

    Net Daily Total Return: This series approximates the minimum possible reinvestment
    of regular cash distributions. Provided that the regular capital repayment is not subject to withholding tax,
    the reinvestment in the Net Daily Total Return is free of withholding tax.

    The Total Return Index that represents the weighted return of the MSCI parent index
    and the cash component.

    The Excess Return Index that represents the return of the Total Return Index
    minus the return of the cash component.
    """

    def __new__(cls, value: str, description: str):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = str(description)
        return obj

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}: " \
               f"'{self.value}' ('{self.description}')>"

    PRICE = ('0', 'Price')
    NET = ('41', 'Net')
    GROSS = ('40', 'Gross')
    TOTAL_RETURN = ('51', 'TR (for Risk Control indexes)')
    EXCESS_RETURN = ('53', 'ER (for Risk Control indexes)')


class Frequencies(enum.Enum):
    """ Index values frequency.
    """
    DAILY = 'D'
    MONTHLY = 'M'
    YEARLY = 'Y'


class Styles(enum.Enum):
    """ Index style.
    """

    def __new__(cls, value: str, description: str):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = str(description)
        return obj

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}: " \
               f"'{self.value}' ('{self.description}')>"

    NONE = ('C', 'None')
    GROWTH = ('G', 'Growth')
    VALUE = ('V', 'Value')


class Sizes(enum.Enum):
    """ Index size.
    """

    def __new__(cls, value: str, description: str, scopes: typing.FrozenSet[Scopes]):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = str(description)
        obj.scopes = frozenset(scopes)
        return obj

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}: " \
               f"'{self.value}' ('{self.description}', {{{','.join(s.description for s in self.scopes)}}})>"

    A_SERIES = ('111', 'A-Series', ())
    REGIONAL_ALL_CAP = ('77', 'All Cap (Large+Mid+Small+Micro Cap)', {Scopes.REGIONAL})
    REGIONAL_ALL_MARKET = ('108', 'All Market', {Scopes.REGIONAL})
    REGIONAL_IMI = ('41', 'IMI (Large+Mid+Small Cap)', {Scopes.REGIONAL})
    REGIONAL_LARGE_CAP = ('37', 'Large Cap', {Scopes.REGIONAL})
    REGIONAL_MICRO_CAP = ('76', 'Micro Cap', {Scopes.REGIONAL})
    REGIONAL_MID_CAP = ('38', 'Mid Cap', {Scopes.REGIONAL})
    PROVISIONAL_IMI = ('119', 'Provisional IMI', ())
    PROVISIONAL_SMALL_CAP = ('99', 'Provisional Small Cap', ())
    PROVISIONAL_STANDARD = ('29', 'Provisional Standard', ())
    REGIONAL_SMID = ('40', 'SMID (Small+Mid Cap)', {Scopes.REGIONAL})
    REGIONAL_SMALL_PLUS_MICRO_CAP = ('79', 'Small + Micro Cap', {Scopes.REGIONAL})
    REGIONAL_SMALL_CAP = ('39', 'Small Cap', {Scopes.REGIONAL})
    REGIONAL_STANDARD = ('36', 'Standard (Large+Mid Cap)', {Scopes.REGIONAL})

    COUNTRY_ALL_CAP = ('75', 'All Cap (Large+Mid+Small+Micro Cap)', {Scopes.COUNTRY})
    COUNTRY_ALL_MARKET = ('107', 'All Market', {Scopes.COUNTRY})
    COUNTRY_IMI = ('35', 'IMI (Large+Mid+Small Cap)', {Scopes.COUNTRY})
    COUNTRY_LARGE_CAP = ('31', 'Large Cap', {Scopes.COUNTRY})
    COUNTRY_MICRO_CAP = ('74', 'Micro Cap', {Scopes.COUNTRY})
    COUNTRY_MID_CAP = ('32', 'Mid Cap', {Scopes.COUNTRY})
    COUNTRY_SMID = ('34', 'SMID (Small+Mid Cap)', {Scopes.COUNTRY})
    COUNTRY_SMALL_PLUS_MICRO_CAP = ('78', 'Small + Micro Cap', {Scopes.COUNTRY})
    COUNTRY_SMALL_CAP = ('33', 'Small Cap', {Scopes.COUNTRY})
    COUNTRY_STANDARD = ('30', 'Standard (Large+Mid Cap)', {Scopes.COUNTRY})


class IndexSuiteGroups(enum.Enum):
    """ Group of index suites.
    """
    CAPPED = 'Capped'
    DOMESTIC = 'Domestic'
    EQUAL_SECTOR_WEIGHTED = 'Equal Sector Weighted'
    EQUAL_COUNTRY_WEIGHTED = 'Equal Country Weighted'
    ESG = 'ESG'
    FACTOR_HIGH_EXPOSURE = 'Factor-High Exposure'
    FACTOR_HIGH_CAPACITY = 'Factor-High Capacity'
    HEDGED = 'Hedged'
    LISTED_REAL_ESTATE = 'Listed Real Estate'
    MULTI_FACTOR = 'Multi-Factor'
    THEMATIC = 'Thematic'
    WMA_PRIVATE_INVESTOR_INDICES = 'WMA Private Investor Indices'


class IndexSuites(enum.Enum):
    """ Index suite.

    See https://app2.msci.com/products/service/index/indexmaster/indexsuites
    """

    def __new__(cls, value: str, description: str, group: typing.Optional[IndexSuiteGroups]):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = str(description)
        obj.group = group
        return obj

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}: " \
               f"'{self.value}' ('{self.description}', {self.group.value!r})>"

    NONE = ('C', 'None', None)

    # Capped
    CAPPED_10_TO_40 = ('AA', '10/40', IndexSuiteGroups.CAPPED)
    CAPPED_25_TO_50 = ('2', '25/50', IndexSuiteGroups.CAPPED)
    STANDARD_CAPPED = ('9', 'Standard Capped', IndexSuiteGroups.CAPPED)

    # Domestic
    HONG_KONG_MPF_DOMESTIC = ('6', 'Hong Kong MPF Domestic', IndexSuiteGroups.DOMESTIC)
    HONG_KONG_MPF_HEDGED = ('8', 'Hong Kong MPF Hedged', IndexSuiteGroups.DOMESTIC)
    HONG_KONG_MPF_UNHEDGED = ('7', 'Hong Kong MPF Unhedged', IndexSuiteGroups.DOMESTIC)

    # Equal Sector Weighted
    EQUAL_SECTOR_WEIGHTED = ('ES', 'Equal Sector Weighted', IndexSuiteGroups.EQUAL_SECTOR_WEIGHTED)

    # Equal Country Weighted
    EQUAL_COUNTRY_WEIGHTED = ('EC', 'Equal Country Weighted', IndexSuiteGroups.EQUAL_COUNTRY_WEIGHTED)

    # ESG
    COUNTRY_ESG_LEADERS = ('CA', 'Country ESG LEADERS', IndexSuiteGroups.ESG)
    ESG_CUSTOM = ('E', 'ESG Custom', IndexSuiteGroups.ESG)
    ESG_FOCUS = ('EF', 'ESG Focus', IndexSuiteGroups.ESG)
    ESG_LEADERS = ('B', 'ESG LEADERS', IndexSuiteGroups.ESG)
    ESG_SCREENED = ('SR', 'ESG Screened', IndexSuiteGroups.ESG)
    ESG_UNIVERSAL = ('EU', 'ESG Universal', IndexSuiteGroups.ESG)
    EMPOWERING_WOMEN = ('EW', 'Empowering Women (WIN)', IndexSuiteGroups.ESG)
    ENVIRONMENTAL = ('Z', 'Environmental', IndexSuiteGroups.ESG)
    EX_CONTROVERSIAL_WEAPONS = ('X', 'Ex Controversial Weapons', IndexSuiteGroups.ESG)
    EX_TOBACCO_INVOLVEMENT = ('TB', 'Ex Tobacco Involvement', IndexSuiteGroups.ESG)
    SRI = ('J', 'SRI', IndexSuiteGroups.ESG)
    WOMENS_LEADERSHIP = ('WL', "Women's Leadership", IndexSuiteGroups.ESG)

    # Factor-High Exposure
    BARRA_FACTOR = ('R', 'Barra Factor', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    BUYBACK_YIELD = ('BY', 'Buyback Yield', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    DIVIDEND_MASTERS = ('DM', 'Dividend Masters', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    ENHANCED_VALUE = ('EV', 'Enhanced Value', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    EQUAL_WEIGHTED = ('W', 'Equal Weighted', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    EQUAL_WEIGHTED_BUYBACK_YIELD = ('EY', 'Equal Weighted Buyback Yield', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    FACTOR_ESG = ('FE', 'Factor ESG', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    GDP_WEIGHTED = ('D', 'GDP Weighted', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    GOVERNANCE_QUALITY = ('GQ', 'Governance-Quality', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    HIGH_DIVIDEND_YIELD = ('H', 'High Dividend Yield', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    MARKET_NEUTRAL = ('3', 'Market Neutral', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    MINIMUM_VOLATILITY = ('M', 'Minimum Volatility', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    MOMENTUM = ('1', 'Momentum', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    FACTOR_HIGH_EXPOSURE_OTHER = ('FO', 'Other', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    PRIME_VALUE = ('PV', 'Prime Value', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    QUALITY = ('U', 'Quality', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    RISK_CONTROL = ('P', 'Risk Control', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    RISK_WEIGHTED = ('K', 'Risk Weighted', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    SECTOR_NEUTRAL_QUALITY = ('NQ', 'Sector Neutral Quality', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    TOP_50_DIVIDEND = ('TD', 'Top 50 Dividend', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)
    TOTAL_SHAREHOLDER_YIELD = ('TS', 'Total Shareholder Yield', IndexSuiteGroups.FACTOR_HIGH_EXPOSURE)

    # Factor-High Capacity
    DIVIDEND_TILT = ('DT', 'Dividend Tilt', IndexSuiteGroups.FACTOR_HIGH_CAPACITY)
    MOMENTUM_TILT = ('MT', 'Momentum Tilt', IndexSuiteGroups.FACTOR_HIGH_CAPACITY)
    QUALITY_TILT = ('QT', 'Quality Tilt', IndexSuiteGroups.FACTOR_HIGH_CAPACITY)
    SIZE_TILT = ('ST', 'Size Tilt', IndexSuiteGroups.FACTOR_HIGH_CAPACITY)
    VALUE_WEIGHTED = ('L', 'Value Weighted', IndexSuiteGroups.FACTOR_HIGH_CAPACITY)
    VOLATILITY_TILT = ('VT', 'Volatility Tilt', IndexSuiteGroups.FACTOR_HIGH_CAPACITY)

    # Hedged
    HEDGED = ('5', 'Hedged', IndexSuiteGroups.HEDGED)
    FACTOR_HEDGED = ('FH', 'Factor Hedged', IndexSuiteGroups.HEDGED)
    ADAPTIVE_HEDGED = ('AH', 'Adaptive Hedged', IndexSuiteGroups.HEDGED)
    ESG_HEDGED = ('EH', 'ESG Hedged', IndexSuiteGroups.HEDGED)

    # Listed Real Estate
    CORE_REAL_ESTATE = ('CE', 'Core Real Estate', IndexSuiteGroups.LISTED_REAL_ESTATE)
    CORE_REAL_ESTATE_FACTOR = ('CF', 'Core Real Estate Factor', IndexSuiteGroups.LISTED_REAL_ESTATE)

    # Multi-Factor
    CUSTOM_FACTOR_MIX = ('CM', 'Custom Factor Mix', IndexSuiteGroups.MULTI_FACTOR)
    FACTOR_MIX_A_SERIES = ('4', 'Factor Mix A-Series', IndexSuiteGroups.MULTI_FACTOR)
    FACTOR_MIX_A_SERIES_CAPPED = ('MA', 'Factor Mix A-Series Capped', IndexSuiteGroups.MULTI_FACTOR)
    DIVERSIFIED_MULTIPLE_FACTOR = ('DF', 'Diversified Multiple-Factor', IndexSuiteGroups.MULTI_FACTOR)
    DIVERSIFIED_FACTOR_MIX = ('FM', 'Diversified Factor Mix', IndexSuiteGroups.MULTI_FACTOR)
    DIVERSIFIED_MULTIPLE_FACTOR_R_SERIES = ('MR', 'Diversified Multiple-Factor R-Series', IndexSuiteGroups.MULTI_FACTOR)
    DIVERSIFIED_MULTIPLE_FACTOR_LOW_VOLATILITY = ('MV', 'Diversified Multiple-Factor Low Volatility',
                                                  IndexSuiteGroups.MULTI_FACTOR)
    DIVERSIFIED_MULTIPLE_5_FACTOR = ('M5', 'Diversified Multiple 5-Factor', IndexSuiteGroups.MULTI_FACTOR)
    DIVERSIFIED_MULTIPLE_3_FACTOR = ('M3', 'Diversified Multiple 3-Factor', IndexSuiteGroups.MULTI_FACTOR)
    ADAPTIVE_MULTIPLE_FACTOR = ('AM', 'Adaptive Multiple Factor', IndexSuiteGroups.MULTI_FACTOR)
    MULTI_FACTOR_OTHER = ('MO', 'Other', IndexSuiteGroups.MULTI_FACTOR)

    # Thematic
    AGRICULTURE_AND_FOOD_CHAIN = ('V', 'Agriculture & Food Chain', IndexSuiteGroups.THEMATIC)
    AGRICULTURE_AND_FOOD_CHAIN_SECTOR_CAPPED = ('Q', 'Agriculture & Food Chain Sector Capped',
                                                IndexSuiteGroups.THEMATIC)
    COMMODITY_PRODUCERS = ('O', 'Commodity Producers', IndexSuiteGroups.THEMATIC)
    COMMODITY_PRODUCERS_SECTOR_CAPPED = ('Y', 'Commodity Producers Sector Capped', IndexSuiteGroups.THEMATIC)
    CONSUMER_DEMAND = ('AB', 'Consumer Demand', IndexSuiteGroups.THEMATIC)
    CONSUMER_GROWTH = ('CG', 'Consumer Growth', IndexSuiteGroups.THEMATIC)
    CYCLICAL_SECTORS = ('CS', 'Cyclical Sectors', IndexSuiteGroups.THEMATIC)
    CYCLICAL_SECTORS_CAPPED = ('CC', 'Cyclical Sectors Capped', IndexSuiteGroups.THEMATIC)
    DEFENSIVE_SECTORS = ('DS', 'Defensive Sectors', IndexSuiteGroups.THEMATIC)
    DEFENSIVE_SECTORS_CAPPED = ('DC', 'Defensive Sectors Capped', IndexSuiteGroups.THEMATIC)
    ECONOMIC_EXPOSURE = ('N', 'Economic Exposure', IndexSuiteGroups.THEMATIC)
    FAITH_BASED = ('F', 'Faith-based', IndexSuiteGroups.THEMATIC)
    INFRASTRUCTURE = ('S', 'Infrastructure', IndexSuiteGroups.THEMATIC)
    INFRASTRUCTURE_CAPPED = ('IC', 'Infrastructure Capped', IndexSuiteGroups.THEMATIC)
    ISLAMIC = ('I', 'Islamic', IndexSuiteGroups.THEMATIC)
    ISLAMIC_M_SERIES_NEW = ('IM', 'Islamic M -Series (New)', IndexSuiteGroups.THEMATIC)
    THEMATIC_OTHER = ('TO', 'Other', IndexSuiteGroups.THEMATIC)

    # WMA Private Investor Indices
    WMA_PRIVATE_INVESTOR_INDICES = ('WM', 'WMA Private Investor Indices',
                                    IndexSuiteGroups.WMA_PRIVATE_INVESTOR_INDICES)


@dataclasses.dataclass
class Context:
    """ Contains the group of parameters that identifies MSCI index apart from its ID.

    This context used by REST API even if we know exact index ID,
    i.e. context is ambiguous but mandatory.
    """
    style: Styles
    size: Sizes
    scope: Scopes

    def __init__(self, *, style: Styles, size: Sizes, scope: Scopes):
        self.style = Styles(style)  # pylint: disable=no-value-for-parameter
        self.size = Sizes(size)  # pylint: disable=no-value-for-parameter
        self.scope = Scopes(scope)  # pylint: disable=no-value-for-parameter


@dataclasses.dataclass
class IndexValue(InstrumentValueProvider):
    """ Container for index history item.
    """
    date: datetime.date
    value: decimal.Decimal
    index_name: str
    style: Styles
    size: Sizes

    def __init__(self,
                 *,
                 date: datetime.date,
                 value: decimal.Decimal,
                 index_name: str,
                 style: Styles,
                 size: Sizes):
        if not isinstance(date, datetime.date):
            raise TypeError("value_date is not date")

        self.date = date
        self.value = decimal.Decimal(value)

        self.index_name = str(index_name)
        self.style = Styles(style)  # pylint: disable=no-value-for-parameter
        self.size = Sizes(size)  # pylint: disable=no-value-for-parameter

    def get_instrument_value(self, tzinfo: typing.Optional[datetime.timezone]) -> InstrumentValue:
        return InstrumentValue(
            value=self.value,
            moment=datetime.datetime.combine(self.date, datetime.time.min, tzinfo=tzinfo))


@dataclasses.dataclass
class IndexInfo(InstrumentInfoProvider):
    """ Container for index information.
    """
    index_id: str
    name: str

    def __init__(self, *, index_id: str, name: str):
        self.index_id = str(index_id)
        self.name = str(name)

    def __str__(self):
        return (f"MSCI index ("
                f"index_id={self.index_id}, "
                f"name={self.name})")

    @property
    def instrument_info(self) -> InstrumentInfo:
        return InstrumentInfo(code=self.index_id, name=self.name)


class MsciIndexesInfoDownloadParameters(typing.NamedTuple):
    """ Container for ``MsciStringDataDownloader.download_instruments_info_string`` parameters.
    """
    market: Markets
    context: Context

    @classmethod
    def safe_create(
            cls: typing.Type['MsciIndexesInfoDownloadParameters'],
            *,
            market: Markets,
            context: Context) -> 'MsciIndexesInfoDownloadParameters':
        """ Create new instance of ``MsciIndexesInfoDownloadParameters`` with arguments check.

        :param market: Market.
        :param context: Context.
        :return: ``MsciIndexesInfoDownloadParameters`` instance.
        """
        if not isinstance(context, Context):
            raise TypeError(f"'context' is not Context: {context!r}")

        return cls(
            market=Markets(market),  # pylint: disable=no-value-for-parameter
            context=context)


@dataclasses.dataclass
class MsciIndexHistoryDownloadParameters(InstrumentHistoryDownloadParameters):
    """ Container for ``MsciStringDataDownloader.download_instrument_history_string`` parameters.
    """
    index_id: Annotated[str, InstrumentInfoParameter(instrument_identity=True)]
    context: Annotated[Context, InstrumentInfoParameter()]
    index_level: IndexLevels
    currency: Currencies
    date_from: Annotated[datetime.date,
                         Description(description="Minimum date of interval to download data. "
                                                 "Usually it equals to the first date of instrument history.")]
    date_to: Annotated[datetime.date,
                       Description(description="Maximum date of interval to download data. "
                                               "It have to be 'today' or '31-12-9999'."),
                       Volatile(generator=lambda ctx: datetime.date.today(), stub_value=datetime.date.max)]

    def clone_with_instrument_info_parameters(
            self,
            info_download_parameters: typing.Optional[MsciIndexesInfoDownloadParameters],
            instrument_info: typing.Optional[IndexInfo]) -> 'MsciIndexHistoryDownloadParameters':
        return MsciIndexHistoryDownloadParameters.generate_from(self, info_download_parameters, instrument_info)

    @classmethod
    def generate_from(
            cls: typing.Type['MsciIndexHistoryDownloadParameters'],
            history_download_parameters: typing.Optional['MsciIndexHistoryDownloadParameters'],
            info_download_parameters: typing.Optional[MsciIndexesInfoDownloadParameters],
            instrument_info: typing.Optional[IndexInfo]) -> 'MsciIndexHistoryDownloadParameters':
        """ Create new history download parameters instance with data from its arguments.

        :param history_download_parameters: Optional instrument history download parameters for cloning.
        :param info_download_parameters: Optional instrument info download parameters for cloning.
        :param instrument_info: Optional instrument info for cloning.
        :return: Cloned history download parameters instance (self) with replacing some attributes from
            `info_download_parameters` and `instrument_info`.
        """
        return cls(
            index_id=((None if history_download_parameters is None else history_download_parameters.index_id)
                      if instrument_info is None
                      else instrument_info.index_id),
            context=((None if history_download_parameters is None else history_download_parameters.context)
                     if info_download_parameters is None
                     else info_download_parameters.context),
            index_level=(None if history_download_parameters is None else history_download_parameters.index_level),
            currency=(None if history_download_parameters is None else history_download_parameters.currency),
            date_from=(None if history_download_parameters is None else history_download_parameters.date_from),
            date_to=(None if history_download_parameters is None else history_download_parameters.date_to),
        )

    @classmethod
    def safe_create(
            cls: typing.Type['MsciIndexHistoryDownloadParameters'],
            *,
            index_id: str,
            context: Context,
            index_level: IndexLevels,
            currency: Currencies,
            date_from: datetime.date,
            date_to: datetime.date) -> 'MsciIndexHistoryDownloadParameters':
        """ Create new instance of ``MsciIndexHistoryDownloadParameters`` with arguments check.

        :param index_id: Index ID.
        :param context: Context/
        :param index_level: Index level.
        :param currency: Currency.
        :param date_from: Download interval beginning.
        :param date_to: Download interval ending.
        :return: ``MsciIndexHistoryDownloadParameters`` instance.
        """
        if not isinstance(context, Context):
            raise TypeError(f"'context' is not Context: {context!r}")
        if not isinstance(date_from, datetime.date):
            raise TypeError(f"'date_from' is not datetime.date: {date_from!r}")
        if not isinstance(date_to, datetime.date):
            raise TypeError(f"'date_to' is not datetime.date: {date_to!r}")
        if date_from > date_to:
            raise ValueError(f"'date_from' ({date_from.isoformat()}) is greater than 'date_to' ({date_to.isoformat()})")

        return cls(
            index_id=str(index_id),
            context=context,
            index_level=IndexLevels(index_level),  # pylint: disable=no-value-for-parameter
            currency=Currencies(currency),  # pylint: disable=no-value-for-parameter
            date_from=date_from,
            date_to=date_to)


class MsciDownloadParametersFactory(DownloadParametersFactory):
    """ Download parameters factories and generators for MSCI.
    """

    @property
    def download_history_parameters_class(self) -> typing.Type[MsciIndexHistoryDownloadParameters]:
        return MsciIndexHistoryDownloadParameters

    @property
    def download_history_parameters_factory(self) -> typing.Callable[..., MsciIndexHistoryDownloadParameters]:
        return MsciIndexHistoryDownloadParameters.safe_create

    @property
    def download_info_parameters_class(self):
        return MsciIndexesInfoDownloadParameters

    @property
    def download_info_parameters_factory(self) -> typing.Callable[..., typing.Any]:
        return MsciIndexesInfoDownloadParameters.safe_create

    def generate_history_download_parameters_from(
            self,
            history_download_parameters: typing.Optional[MsciIndexHistoryDownloadParameters],
            info_download_parameters: typing.Optional[MsciIndexesInfoDownloadParameters],
            instrument_info: typing.Optional[IndexInfo]) -> MsciIndexHistoryDownloadParameters:
        return MsciIndexHistoryDownloadParameters.generate_from(
            history_download_parameters,
            info_download_parameters,
            instrument_info)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Utilities for parse data from app2.msci.com
"""

import decimal
import json
import logging
import typing
import datetime

from .meta import (
    IndexValue, IndexInfo, Market, Size, Style, IndexSuite, IndexSuiteGroup, Frequency,
    Scopes, IndexLevel, Currency, IndexPanelData)
from ...base import (
    InstrumentValuesHistoryParser, InstrumentInfoParser, ParseError, SourceDownloadError,
    DownloadParameterValuesStorage, InstrumentValuesHistoryEmpty)

logging.getLogger().addHandler(logging.NullHandler())


class MsciHistoryJsonParser(InstrumentValuesHistoryParser):
    """ Parser for history data of index from JSON string.

    https://app2.msci.com/products/service/index/indexmaster/getLevelDataForGraph?currency_symbol=USD&index_variant=STRD&start_date=20170813&end_date=20210813&data_frequency=DAILY&index_codes=990300

    E.g.::

        {
            "msci_index_code":"990300",
            "index_variant_type":"STRD",
            "ISO_currency_symbol":"USD",
            "indexes":{
                "INDEX_LEVELS":[
                    {"level_eod":1892.970699,"calc_date":20170609},
                    {"level_eod":1886.335805,"calc_date":20170612},
                    ...
                    ]
            }
        }

    On error may return

    ::

     {
      "timestamp":"Aug 14, 2021 4:20:12 PM",
      "user":"sys_x_bmapip01",
      "error_code":" 100",
      "error_message":" null Invalid Parameter end_date : '19691231. Calculation date cannot be earlier than 19970101'"
     }

    Or

    ::

        {
            "timestamp":"Aug 14, 2021 4:23:16 PM",
            "user":"sys_x_bmapip01",
            "error_code":"  300",
            "error_message":"  Invalid Parameter data_frequency : 'DAYLY'"
        }
    """
    date_format = '%Y%m%d'

    def __init__(self, parameter_values_storage: DownloadParameterValuesStorage):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.parameter_values_storage = parameter_values_storage

    def parse(  # pylint: disable=arguments-renamed
            self,
            raw_json_text: str,
            tzinfo: typing.Optional[datetime.timezone]
    ) -> typing.Iterable[IndexValue]:

        try:
            raw_data = json.loads(raw_json_text)
        except json.decoder.JSONDecodeError as ex:
            raise ParseError(ex.msg) from ex

        if not isinstance(raw_data, dict):
            # can be Inf etc., but we accept only {}
            raise ParseError("Wrong JSON format. Top level is not dictionary.")

        error_message = raw_data.get('error_message', None)
        if error_message is not None:
            raise SourceDownloadError(f"JSON contains error message: {error_message}")

        msci_index_code = raw_data.get('msci_index_code', None)
        if msci_index_code is None:
            raise ParseError("Wrong JSON format. 'msci_index_code' not found.")

        index_variant_type = raw_data.get('index_variant_type', None)
        if index_variant_type is None:
            raise ParseError("Wrong JSON format. 'index_variant_type' not found.")

        iso_currency_symbol = raw_data.get('ISO_currency_symbol', None)
        if iso_currency_symbol is None:
            raise ParseError("Wrong JSON format. 'ISO_currency_symbol' not found.")

        indexes_block = raw_data.get('indexes', None)
        if indexes_block is None:
            raise ParseError("Wrong JSON format. 'indexes' block not found.")

        if not isinstance(indexes_block, dict):
            raise ParseError("Wrong JSON format. 'indexes' block is not dictionary.")

        index_levels_block = indexes_block.get('INDEX_LEVELS', None)
        if index_levels_block is None:
            raise ParseError("Wrong JSON format. 'INDEX_LEVELS' block not found.")

        if not index_levels_block:
            raise InstrumentValuesHistoryEmpty()

        index_values_pairs: typing.List[typing.Tuple[decimal.Decimal, datetime.date]] = []
        for index_value_block in index_levels_block:
            if not isinstance(index_value_block, dict):
                raise ParseError("Wrong JSON format. Item inside 'INDEX_LEVELS' block is not dictionary.")

            level_eod = index_value_block.get('level_eod', None)
            if level_eod is None:
                raise ParseError("Wrong JSON format. 'level_eod' not found.")

            calc_date = index_value_block.get('calc_date', None)
            if calc_date is None:
                raise ParseError("Wrong JSON format. 'calc_date' not found.")

            if isinstance(level_eod, float):
                # hack to adjust floating point digits
                level_eod = repr(level_eod)

            try:
                level_eod = decimal.Decimal(level_eod)
            except (ValueError, TypeError, decimal.DecimalException) as ex:
                raise ParseError(f"Wrong JSON format. Can't convert {level_eod} to decimal.") from ex

            try:
                calc_date = datetime.datetime.strptime(str(calc_date), self.date_format)
            except (ValueError, TypeError) as ex:
                raise ParseError(f"Wrong JSON format. Can't convert {calc_date} to datetime.") from ex

            index_values_pairs.append((level_eod, calc_date.date()))

        index_level = self.parameter_values_storage.get_dynamic_enum_value_by_key(IndexLevel, index_variant_type)
        if index_level is None:
            raise ParseError(f"Index level {index_variant_type!r} not found")

        currency = self.parameter_values_storage.get_dynamic_enum_value_by_key(Currency, iso_currency_symbol)
        if currency is None:
            raise ParseError(f"Currency {iso_currency_symbol!r} not found")

        return (
            IndexValue(calc_date=calc_date,
                       level_eod=level_eod,
                       msci_index_code=msci_index_code,
                       index_variant_type=index_level,
                       currency=currency)
            for level_eod, calc_date
            in index_values_pairs)


class MsciIndexInfoParser(InstrumentInfoParser):
    """ Parser for indexes info list from JSON.

    https://app2.msci.com/products/service/index/indexmaster/searchIndexes?index_market=24576&index_scope=Region&index_size=12&index_style=None&index_suite=C

    E.g.::

        {
            "indexes":[
                {"msci_index_code":903600,"index_name":"AUSTRALIA"},
                {"msci_index_code":904000,"index_name":"AUSTRIA"},
                ...
                ]
        }
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

    def parse(self, raw_json_text: str) -> typing.Iterable[IndexInfo]:  # pylint: disable=arguments-renamed
        try:
            raw_data = json.loads(raw_json_text)
        except json.decoder.JSONDecodeError as ex:
            raise ParseError(ex.msg) from ex

        if not isinstance(raw_data, dict):
            # can be Inf etc., but we accept only {}
            raise ParseError("Wrong JSON format. Top level is not dictionary.")

        indexes_block = raw_data.get('indexes', None)
        if indexes_block is None:
            raise ParseError("Wrong JSON format. 'indexes' block not found.")

        index_info_pairs: typing.List[typing.Tuple[str, str]] = []
        for index_info_block in indexes_block:
            if not isinstance(index_info_block, dict):
                raise ParseError("Wrong JSON format. Item inside 'indexes' block is not dictionary.")

            msci_index_code = index_info_block.get('msci_index_code', None)
            if msci_index_code is None:
                raise ParseError("Wrong JSON format. 'msci_index_code' not found.")

            index_name = index_info_block.get('index_name', None)
            if index_name is None:
                raise ParseError("Wrong JSON format. 'index_name' not found.")

            index_info_pairs.append((msci_index_code, index_name))

        return (
            IndexInfo(msci_index_code=msci_index_code,
                      index_name=index_name)
            for msci_index_code, index_name
            in index_info_pairs)


class MsciIndexPanelDataJsonParser:
    """ Parser for index panel data from JSON.
    """

    all_country_market_id = '24576'
    daily_frequency_id = 'DAILY'
    monthly_frequency_id = 'END_OF_MONTH'

    def __init__(self):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self._index_suite_groups = {}

    @staticmethod
    def _parse_block(
            block_name: str,
            src_block: typing.Dict[str, typing.List]) -> typing.Iterable[typing.Tuple[str, str]]:

        target_block = src_block.get(block_name, None)
        if target_block is None:
            raise ParseError(f"Wrong JSON format. {block_name!r} block not found.")

        for pair_block in target_block:
            if not isinstance(pair_block, dict):
                raise ParseError(f"Wrong JSON format. Item inside {block_name!r} block is not dictionary.")

            identity = pair_block.get('id', None)
            if identity is None:
                raise ParseError("Wrong JSON format. 'id' not found.")

            name = pair_block.get('name', None)
            if name is None:
                raise ParseError("Wrong JSON format. 'name' not found.")

            yield identity, name

    def parse(self, raw_json_text: str) -> IndexPanelData:
        """ Parse index panel data from JSON and return it.

        :param raw_json_text: JSON string with index panel data.
        :return: ``IndexPanelData`` instance.
        """
        try:
            raw_data = json.loads(raw_json_text)
        except json.decoder.JSONDecodeError as ex:
            raise ParseError(ex.msg) from ex

        if not isinstance(raw_data, dict):
            # can be Inf etc., but we accept only {}
            raise ParseError("Wrong JSON format. Top level is not dictionary.")

        markets = tuple(
            Market.safe_create(
                identity=identity,
                name=name,
                # 'All Country (DM+EM)' market available only in regional scope
                scope=Scopes.REGIONAL if str(identity) == self.all_country_market_id else None)
            for identity, name
            in self._parse_block('market', raw_data))
        currencies = tuple(
            Currency.safe_create(identity=identity, name=name)
            for identity, name
            in self._parse_block('currency', raw_data))
        index_levels = tuple(
            IndexLevel.safe_create(identity=identity, name=name)
            for identity, name
            in self._parse_block('indexLevels', raw_data))
        frequencies = tuple(
            Frequency.safe_create(identity=identity, name=name)
            for identity, name
            in self._parse_block('frequency', raw_data))
        sizes = tuple(
            Size.safe_create(identity=identity, name=name)
            for identity, name
            in self._parse_block('size', raw_data))
        styles = tuple(
            Style.safe_create(identity=identity, name=name)
            for identity, name
            in self._parse_block('style', raw_data))

        index_suites_block = raw_data.get('indexSuite', None)
        if index_suites_block is None:
            raise ParseError("Wrong JSON format. 'indexSuite' block not found.")

        index_suite_groups: typing.List[IndexSuiteGroup] = []
        index_suites: typing.List[IndexSuite] = []

        for index_suites_block_item in index_suites_block:
            optgroup = index_suites_block_item.get('optgroup', None)
            if optgroup is None:
                index_suites.extend(
                    IndexSuite.safe_create(identity=identity, name=name)
                    for identity, name
                    in self._parse_block('DUMMY', {'DUMMY': [index_suites_block_item]}))

            else:
                group = IndexSuiteGroup.safe_create(name=optgroup)
                index_suite_groups.append(group)
                index_suites.extend(
                    IndexSuite.safe_create(identity=identity, name=name, group=group)
                    for identity, name
                    in self._parse_block('options', index_suites_block_item))

        # unfortunately, daily and monthly frequencies hard-coded in site scripts
        # so we forced to assume their codes here
        frequencies_dict = {frequency.identity: frequency for frequency in frequencies}
        if self.daily_frequency_id not in frequencies_dict:
            raise ParseError(f"Frequency with id {self.daily_frequency_id!r} not found.")
        if self.monthly_frequency_id not in frequencies_dict:
            raise ParseError(f"Frequency with id {self.monthly_frequency_id!r} not found.")

        return IndexPanelData(
            markets=markets,
            currencies=currencies,
            index_levels=index_levels,
            frequencies=frequencies,
            index_suite_groups=tuple(index_suite_groups),
            index_suites=tuple(index_suites),
            sizes=sizes,
            styles=styles,
            daily_frequency=frequencies_dict[self.daily_frequency_id],
            monthly_frequency=frequencies_dict[self.monthly_frequency_id])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import decimal
import unittest

from sane_finances.sources.base import (
    InstrumentValue, InstrumentInfo, DownloadParametersFactory, InstrumentHistoryDownloadParameters)
from sane_finances.sources.moex.v1_3.meta import (
    Market, TradeEngine, Board,
    SecurityValue, SecurityInfo, MoexSecurityHistoryDownloadParameters, MoexSecuritiesInfoDownloadParameters,
    MoexDownloadParametersFactory)
from .common import CommonTestCases


class TestTradeEngine(unittest.TestCase):

    def test_safe_create_Success(self):
        _ = TradeEngine.safe_create(identity=42, name='NAME', title='TITLE')


class TestMarket(unittest.TestCase):

    def test_safe_create_Success(self):
        _ = Market.safe_create(
            identity=42,
            trade_engine=TradeEngine(identity=42, name='NAME', title='TITLE'),
            name='NAME',
            title='TITLE',
            marketplace='MARKETPLACE')

    def test_safe_create_raiseWrongEngine(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = Market.safe_create(
                identity=42,
                trade_engine=None,
                name='NAME',
                title='TITLE',
                marketplace='MARKETPLACE')


class TestBoard(unittest.TestCase):

    def setUp(self) -> None:
        self.engine = TradeEngine(identity=42, name='NAME', title='TITLE')
        self.market = Market(
            identity=42,
            trade_engine=self.engine,
            name='NAME',
            title='TITLE',
            marketplace='MARKETPLACE')

    def test_safe_create_Success(self):
        _ = Board.safe_create(
            identity=42,
            trade_engine=self.engine,
            market=self.market,
            boardid='BOARDID',
            title='TITLE',
            is_traded=True,
            has_candles=True,
            is_primary=True)

    def test_safe_create_raiseWrongEngine(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = Board.safe_create(
                identity=42,
                trade_engine=None,
                market=self.market,
                boardid='BOARDID',
                title='TITLE',
                is_traded=True,
                has_candles=True,
                is_primary=True)

    def test_safe_create_raiseWrongMarket(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = Board.safe_create(
                identity=42,
                trade_engine=self.engine,
                market=None,
                boardid='BOARDID',
                title='TITLE',
                is_traded=True,
                has_candles=True,
                is_primary=True)


class TestSecurityValue(unittest.TestCase):

    def test_instrument_value_Success(self):
        security_value = SecurityValue(
            trade_date=datetime.date(2000, 12, 31),
            close=decimal.Decimal(42))
        expected_instrument_value = InstrumentValue(
            value=security_value.close,
            moment=datetime.datetime.combine(security_value.trade_date,
                                             datetime.time.min,
                                             tzinfo=datetime.timezone.utc))

        instrument_value = security_value.get_instrument_value(datetime.timezone.utc)

        self.assertEqual(expected_instrument_value, instrument_value)

    def test_raiseWrongDate(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = SecurityValue(
                trade_date=None,
                close=decimal.Decimal(42))


class TestSecurityInfo(unittest.TestCase):

    def test_instrument_value_Success(self):
        engine = TradeEngine(identity=42, name='NAME', title='TITLE')
        market = Market(
            identity=42,
            trade_engine=engine,
            name='NAME',
            title='TITLE',
            marketplace='MARKETPLACE')

        board = Board.safe_create(
            identity=42,
            trade_engine=engine,
            market=market,
            boardid='BOARDID',
            title='TITLE',
            is_traded=True,
            has_candles=True,
            is_primary=True)

        security_info = SecurityInfo(
            sec_id='ID',
            board=board,
            short_name='SHORT NAME')
        expected_instrument_info = InstrumentInfo(code=security_info.sec_id, name=security_info.short_name)

        instrument_info = security_info.instrument_info

        self.assertEqual(expected_instrument_info, instrument_info)

    def test_instrument_value_RaiseWhenWrongBoard(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = SecurityInfo(
                sec_id='ID',
                board=None,
                short_name='SHORT NAME')


class TestMoexSecuritiesInfoDownloadParameters(unittest.TestCase):

    def test_safe_create_Success(self):
        engine = TradeEngine(identity=42, name='NAME', title='TITLE')
        market = Market(
            identity=42,
            trade_engine=engine,
            name='NAME',
            title='TITLE',
            marketplace='MARKETPLACE')

        board = Board.safe_create(
            identity=42,
            trade_engine=engine,
            market=market,
            boardid='BOARDID',
            title='TITLE',
            is_traded=True,
            has_candles=True,
            is_primary=True)

        _ = MoexSecuritiesInfoDownloadParameters.safe_create(board=board)

    def test_safe_create_raiseWrongBoard(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = MoexSecuritiesInfoDownloadParameters.safe_create(board=None)


class TestMoexSecurityHistoryDownloadParameters(unittest.TestCase):

    def setUp(self) -> None:
        engine = TradeEngine(identity=42, name='NAME', title='TITLE')
        market = Market(
            identity=42,
            trade_engine=engine,
            name='NAME',
            title='TITLE',
            marketplace='MARKETPLACE')

        self.board = Board.safe_create(
            identity=42,
            trade_engine=engine,
            market=market,
            boardid='BOARDID',
            title='TITLE',
            is_traded=True,
            has_candles=True,
            is_primary=True)

    def test_safe_create_Success(self):
        _ = MoexSecurityHistoryDownloadParameters.safe_create(
            board=self.board,
            sec_id='ID',
            start=0)

    def test_safe_create_raiseWrongBiard(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = MoexSecurityHistoryDownloadParameters.safe_create(
                board=None,
                sec_id='ID',
                start=0)

    def test_clone_with_instrument_info_parameters_Success(self):
        params = MoexSecurityHistoryDownloadParameters.safe_create(
            board=self.board,
            sec_id='ID',
            start=0)
        expected_result = params

        result = params.clone_with_instrument_info_parameters(None, None)

        self.assertEqual(expected_result, result)


class TestMoexDownloadParametersFactory(CommonTestCases.CommonDownloadParametersFactoryTests):

    def get_download_parameters_factory(self) -> DownloadParametersFactory:
        return MoexDownloadParametersFactory()

    def get_generate_history_download_parameters_from_expected_result_with_none(self) \
            -> InstrumentHistoryDownloadParameters:
        # noinspection PyTypeChecker
        expected_result = MoexSecurityHistoryDownloadParameters(board=None, sec_id=None, start=0)
        return expected_result



#!/usr/bin/python
# -*- coding: utf-8 -*-

""" Inspection utilities for sources.
"""


class InstrumentInfoParameter:
    """ Describes attribute of instrument's history download parameters class
    which value comes from instrument's info download parameters
    and should not be changed without changing instrument itself.

    For example: some sources provide API for searching instruments by parameters;
    such API returns info about instrument (unique code, name, etc.).
    In spite of this other API (e.g. download values history of instrument) may need same parameters
    in addition to other specific parameters.

    ::

        class InfoParams:
            ''' This params uses for searching of instruments list.
                Such searching returns unique ISIN of instruments.
            '''
            region: str
            country: str

        class HistoryParams:
            ''' Despite of fact that ISIN uniquely identifies concrete instrument
                this API demands region and country codes for instrument values history.
                So we mark them properly.
            '''
            region: typing.Annotated[str, InstrumentInfoParameter()]
            country: typing.Annotated[str, InstrumentInfoParameter()]
            isin: typing.Annotated[str, InstrumentInfoParameter(instrument_identity=True)]
            currency: str
    """
    instrument_identity: bool

    def __init__(self, instrument_identity: bool = False):
        """ Initialize annotation.

        :param instrument_identity: ``True`` if annotated attribute represents instrument identity value.
        """
        self.instrument_identity = bool(instrument_identity)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Tools for download (export) index data from app2.msci.com
"""
import collections
import dataclasses
import logging
import typing
import datetime
import decimal
import inspect

from .meta import (Markets, Formats, Currencies, IndexLevels, Frequencies, Styles, Sizes, Scopes, Context,
                   IndexSuites, IndexSuiteGroups,
                   MsciIndexHistoryDownloadParameters, MsciIndexesInfoDownloadParameters,
                   MsciDownloadParametersFactory)
from .parsers import MsciHistoryXmlParser, MsciIndexInfoParser
from ...base import (
    ApiActualityChecker, InstrumentStringDataDownloader, ParseError,
    InstrumentExporterFactory, InstrumentHistoryValuesExporter, InstrumentsInfoExporter,
    DownloadParameterValuesStorage, CheckApiActualityError)
from ...generic import GenericInstrumentHistoryValuesExporter, GenericInstrumentsInfoExporter
from ....communication.downloader import Downloader, DownloadStringResult

logging.getLogger().addHandler(logging.NullHandler())


class MsciStringDataDownloader(InstrumentStringDataDownloader):
    """ Data downloader from app2.msci.com.

    The list of all instruments can be obtained from
    https://app2.msci.com/webapp/indexperf/pages/IEIPerformanceRegional.jsf

    There inside HTML, we can find all necessary instruments and their available values for REST API.

    Examples of HTML pages for browsing of index history:
    https://app2.msci.com/products/indexes/performance/regional_chart.html?asOf=Feb%2025,%202020&size=36&scope=R&style=C&currency=15&priceLevel=0&indexId=3478
    or
    https://app2.msci.com/products/indexes/performance/country_chart.html?asOf=Mar%2020,%202020&size=30&scope=C&style=C&currency=15&priceLevel=0&indexId=104

    When clicked "Add/Remove indexes" executes request like
    https://app2.msci.com/webapp/indexperf/charts?style=C&market=1896&getIndices=true&size=36&site=gimi&scope=R

    It returns list of indexes by specified parameters in form like
    ``<?xml version="1.0" ?><indices><index id="2670" name="AC AMERICAS" />...</indices>``

    When clicked "Update chart" executes request like
    https://app2.msci.com/webapp/indexperf/charts?endDate=25%20Feb%2C%202020&baseValue=false&frequency=D&currency=15&indices=3478%2CC%2C36&format=XML&site=gimi&scope=R&startDate=25%20Feb%2C%202016&priceLevel=0

    It returns a history of index values by specified parameters in form like
    ::

        <?xml version="1.0" ?>  <performance>
        <index id="EAFE,C,36">
          <asOf>
            <date>03/11/2016</date>
            <value>1,644.941</value>
          </asOf>
          ...
        </index>
      </performance>
    """

    IndexDataUrl = 'https://app2.msci.com/webapp/indexperf/charts'

    def __init__(self, downloader: Downloader):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.downloader = downloader

        # headers for HTTP
        self.headers: typing.Dict[str, str] = {}

        # common requests parameters
        self.get_indices = True
        self.site = 'gimi'
        self.format = Formats.XML
        self.base_value = False
        self.frequency = Frequencies.DAILY

    def adjust_download_instrument_history_parameters(
            self,
            parameters: MsciIndexHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime
    ) -> typing.Tuple[MsciIndexHistoryDownloadParameters, datetime.datetime, datetime.datetime]:

        parameters, moment_from, moment_to = super().adjust_download_instrument_history_parameters(
            parameters, moment_from, moment_to)
        parameters: MsciIndexHistoryDownloadParameters

        date_from = parameters.date_from
        if date_from < moment_from.date():
            # try to shift date_from closer to moment_from
            date_from = moment_from.date()

            parameters = dataclasses.replace(parameters, date_from=date_from)

        date_to = parameters.date_to
        if date_to != moment_to.date():
            # try to shift date_to closer to moment_from
            date_to = moment_to.date()

            parameters = dataclasses.replace(parameters, date_to=date_to)

        return parameters, moment_from, moment_to

    def paginate_download_instrument_history_parameters(
            self,
            parameters: MsciIndexHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime
    ) -> typing.Iterable[typing.Tuple[MsciIndexHistoryDownloadParameters, datetime.datetime, datetime.datetime]]:

        assert parameters.date_from <= parameters.date_to
        assert moment_from <= moment_to

        if moment_from.date() == moment_to.date():
            # if we need data only for one day, than we can't split it
            yield parameters, moment_from, moment_to
            return

        today = datetime.date.today()
        # this API can download daily data only for last 5 years
        five_years_ago = today.replace(year=today.year - 5)  # minimum date when we can get daily data

        if moment_from.date() >= five_years_ago or moment_to.date() < five_years_ago:
            # if we need data only for less than last 5 years,
            # or we don't need data for last 5 years,
            # then there's nothing to do
            yield parameters, moment_from, moment_to
            return

        # otherwise we split interval in two parts:
        # - first part (data older than 5 years) will treat in monthly frequency
        # - second part (last 5 years) will treat in any desired frequency

        first_moment_from = moment_from
        date_to = five_years_ago - datetime.timedelta(days=1)
        first_moment_to = datetime.datetime.combine(date_to, datetime.time.min, tzinfo=moment_to.tzinfo)
        params_first = dataclasses.replace(parameters, date_to=date_to)
        if params_first.date_from > params_first.date_to:
            params_first = dataclasses.replace(params_first, date_from=params_first.date_to)

        second_moment_from = datetime.datetime.combine(five_years_ago, datetime.time.min, tzinfo=moment_from.tzinfo)
        second_moment_to = moment_to
        params_second = dataclasses.replace(parameters, date_from=five_years_ago)
        if params_second.date_from > params_second.date_to:
            params_second = dataclasses.replace(params_second, date_to=params_second.date_from)

        yield params_first, first_moment_from, first_moment_to
        yield params_second, second_moment_from, second_moment_to

    def download_instrument_history_string(
            self,
            parameters: MsciIndexHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime) -> DownloadStringResult:
        date_from = moment_from.date()
        date_to = moment_to.date()
        if moment_to.time() != datetime.time.min:
            date_to += datetime.timedelta(days=1)

        return self.download_index_history_string(
            parameters.index_id,
            parameters.context,
            parameters.index_level,
            parameters.currency,
            date_from,
            date_to)

    def download_instruments_info_string(self, parameters: MsciIndexesInfoDownloadParameters) -> DownloadStringResult:
        return self.download_indexes_info_string(parameters.market, parameters.context)

    def download_index_history_string(
            self,
            index_id: str,
            context: Context,
            index_level: IndexLevels,
            currency: Currencies,
            date_from: datetime.date,
            date_to: datetime.date) -> DownloadStringResult:
        """ Downloads data for one index as string.

        :param index_id: Index ID.
        :param context: Context.
        :param index_level: Index level.
        :param currency: Currency.
        :param date_from: Download interval beginning.
        :param date_to: Download interval ending.
        :return: Container with downloaded string.
        """
        index_param = f"{index_id},{context.style.value},{context.size.value}"

        params = [
            ('format', self.format.value),
            ('site', self.site),
            ('baseValue', str(self.base_value).lower()),
            ('indices', index_param),
            ('scope', context.scope.value),
            ('frequency', self.frequency.value),
            ('currency', currency.value),
            ('priceLevel', index_level.value),
            ('startDate', date_from.strftime("%d %b, %Y")),  # WARNING!!! locale dependent format
            ('endDate', date_to.strftime("%d %b, %Y"))  # WARNING!!! locale dependent format
        ]

        self.downloader.parameters = params
        self.downloader.headers = self.headers

        return self.downloader.download_string(self.IndexDataUrl)

    def download_indexes_info_string(self, market: Markets, context: Context) -> DownloadStringResult:
        """ Downloads the list of all available indexes by specified parameters

        :param market: Market.
        :param context: Context.
        :return: Container with downloaded string.
        """
        self.downloader.headers = {}
        self.downloader.parameters = [
            ('getIndices', str(self.get_indices).lower()),
            ('site', self.site),
            ('market', market.value),
            ('style', context.style.value),
            ('size', context.size.value),
            ('scope', context.scope.value)
        ]

        return self.downloader.download_string(self.IndexDataUrl)


class MsciApiActualityChecker(ApiActualityChecker):
    """ Verifies actuality and accessibility of REST API of app2.msci.com.
    """

    _expectedIndexName = 'USA'
    _expectedIndexContext = Context(style=Styles.NONE, size=Sizes.COUNTRY_STANDARD, scope=Scopes.COUNTRY)
    _expectedFirstDate = datetime.date(1969, 12, 31)
    _expectedFirstValue = decimal.Decimal('100.0')

    def __init__(self,
                 string_data_downloader: MsciStringDataDownloader,
                 history_xml_parser: MsciHistoryXmlParser,
                 index_info_parser: MsciIndexInfoParser):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.string_data_downloader = string_data_downloader
        self.history_xml_parser = history_xml_parser
        self.index_info_parser = index_info_parser

    def check(self):
        self.logger.info("Check actuality via index list")

        index_id = '104'  # USA
        index_market = Markets.COUNTRY_DEVELOPED_MARKETS
        index_context = self._expectedIndexContext
        index_level = IndexLevels.PRICE
        index_currency = Currencies.USD
        index_history_date_from, index_history_date_to = self._expectedFirstDate, datetime.date(2020, 1, 1)

        self.string_data_downloader.get_indices = True
        self.string_data_downloader.format = Formats.XML
        self.string_data_downloader.base_value = False
        self.string_data_downloader.frequency = Frequencies.MONTHLY

        xml_list_result = self.string_data_downloader.download_indexes_info_string(index_market, index_context)
        self.logger.debug(f"Got XML data:\n{xml_list_result.downloaded_string}")

        try:
            info_list = list(self.index_info_parser.parse(xml_list_result.downloaded_string))
        except ParseError as ex:
            xml_list_result.set_correctness(False)
            raise CheckApiActualityError(f"Unexpected index info XML: {ex.message}") from ex
        except Exception:
            xml_list_result.set_correctness(False)
            raise

        if len(info_list) == 0:
            xml_list_result.set_correctness(False)
            raise CheckApiActualityError("Unexpected index info list. No data")

        # here we know that check via index list was successful

        self.logger.info(f"Check actuality via index {index_id!r}")

        xml_history_result = self.string_data_downloader.download_index_history_string(
            index_id,
            index_context,
            index_level,
            index_currency,
            index_history_date_from,
            index_history_date_to
        )

        try:
            history = list(self.history_xml_parser.parse(
                xml_history_result.downloaded_string,
                tzinfo=None))
        except ParseError as ex:
            xml_history_result.set_correctness(False)
            raise CheckApiActualityError(f"Unexpected index history XML: {ex.message}") from ex
        except Exception:
            xml_history_result.set_correctness(False)
            raise

        history.sort(key=lambda v: v.date)

        first_value = history[0]  # for actuality checking one record is enough

        if first_value.index_name != self._expectedIndexName:
            xml_history_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Unexpected index name. {first_value.index_name!r} != {self._expectedIndexName!r}")

        if first_value.size != self._expectedIndexContext.size:
            xml_history_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Unexpected index size. {first_value.size!r} != {self._expectedIndexContext.size!r}")

        if first_value.style != self._expectedIndexContext.style:
            xml_history_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Unexpected index style. {first_value.style} != {self._expectedIndexContext.style}")

        if first_value.date != self._expectedFirstDate:
            xml_history_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Unexpected first date. {first_value.date} != {self._expectedFirstDate}")

        if first_value.value != self._expectedFirstValue:
            xml_history_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Unexpected first value. {first_value.value} != {self._expectedFirstValue}")

        self.logger.info("Actuality check was successful")


class MsciIndexDownloadParameterValuesStorage(DownloadParameterValuesStorage):
    """ Storage of indexes download parameters.
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self._special_handlers: typing.Dict[type, typing.Callable] = {
            Markets: self._get_markets_choices,
            Sizes: self._get_sizes_choices,
            IndexSuites: self._get_index_suites_choices
        }

    def is_dynamic_enum_type(self, cls: type) -> bool:
        return False

    def get_all_managed_types(self) -> typing.Iterable[typing.Type]:
        return ()

    def get_dynamic_enum_key(self, instance):
        return None

    def get_parameter_type_choices(self, cls: type) \
            -> typing.Optional[
                typing.List[typing.Tuple[typing.Any, typing.Union[str, typing.List[typing.Tuple[typing.Any, str]]]]]
            ]:

        if not inspect.isclass(cls):
            return None

        if cls not in self._special_handlers:
            return None

        return self._special_handlers[cls]()

    @staticmethod
    def _get_markets_choices():
        market: Markets

        grouped_markets = collections.OrderedDict()
        for market in Markets:
            if market.scope not in grouped_markets:
                grouped_markets[market.scope] = []
            grouped_markets[market.scope].append(market)

        return [(f"Available for {scope.description!r} scope:",
                 [(market.value, market.description) for market in markets])
                for scope, markets
                in grouped_markets.items()]

    @staticmethod
    def _get_sizes_choices():
        size: Sizes

        grouped_sizes = collections.OrderedDict({None: []})
        for size in Sizes:
            if not size.scopes:  # available for all scopes
                grouped_sizes[None].append(size)
            else:
                for scope in size.scopes:
                    if scope not in grouped_sizes:
                        grouped_sizes[scope] = []
                    grouped_sizes[scope].append(size)

        return [(size.value, size.description) for size in grouped_sizes[None]] + \
               [(f"Available for {scope.description!r} scope:",
                 [(size.value, size.description) for size in sizes])
                for scope, sizes
                in grouped_sizes.items() if scope is not None]

    @staticmethod
    def _get_index_suites_choices():
        suite: IndexSuites
        group: IndexSuiteGroups

        suites_without_groups = [(suite.value, suite.description) for suite in IndexSuites if suite.group is None]
        grouped_suites = [(group.value, [(suite.value, suite.description)
                                         for suite
                                         in IndexSuites
                                         if suite.group == group])
                          for group in IndexSuiteGroups]

        return suites_without_groups + grouped_suites


class MsciIndexExporterFactory(InstrumentExporterFactory):
    """ Factory class for create instances of MSCI data exporter.
    """
    name: str = 'MSCI index data exporter. Version before 2021.'
    provider_site: str = 'https://www.msci.com/'
    api_url: str = 'https://app2.msci.com/webapp/indexperf/pages/IEIPerformanceRegional.jsf'

    def __init__(self):
        self._dynamic_enum_type_manager = None
        self._download_parameters_factory = None

    def create_history_values_exporter(self, downloader: Downloader) -> InstrumentHistoryValuesExporter:
        string_data_downloader = MsciStringDataDownloader(downloader)
        history_values_parser = MsciHistoryXmlParser()

        return GenericInstrumentHistoryValuesExporter(string_data_downloader, history_values_parser)

    def create_info_exporter(self, downloader: Downloader) -> InstrumentsInfoExporter:
        string_data_downloader = MsciStringDataDownloader(downloader)
        info_parser = MsciIndexInfoParser()

        return GenericInstrumentsInfoExporter(string_data_downloader, info_parser)

    def create_download_parameter_values_storage(self, downloader: Downloader) -> DownloadParameterValuesStorage:
        return MsciIndexDownloadParameterValuesStorage()

    def create_api_actuality_checker(self, downloader: Downloader) -> MsciApiActualityChecker:
        string_data_downloader = MsciStringDataDownloader(downloader)
        history_values_parser = MsciHistoryXmlParser()
        info_parser = MsciIndexInfoParser()

        return MsciApiActualityChecker(string_data_downloader, history_values_parser, info_parser)

    @property
    def dynamic_enum_type_manager(self) -> MsciIndexDownloadParameterValuesStorage:
        if self._dynamic_enum_type_manager is None:
            self._dynamic_enum_type_manager = MsciIndexDownloadParameterValuesStorage()

        return self._dynamic_enum_type_manager

    @property
    def download_parameters_factory(self) -> MsciDownloadParametersFactory:
        if self._download_parameters_factory is None:
            self._download_parameters_factory = MsciDownloadParametersFactory()

        return self._download_parameters_factory



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import decimal
import re
import unittest

from sane_finances.sources.base import ParseError, InstrumentValuesHistoryEmpty
from sane_finances.sources.moex.v1_3.meta import (
    Market, TradeEngine, Board,
    SecurityValue, SecurityInfo, GlobalIndexData)
from sane_finances.sources.moex.v1_3.parsers import (
    MoexHistoryJsonParser, MoexSecurityInfoJsonParser, MoexGlobalIndexJsonParser)
from .fakes import FakeMoexDownloadParameterValuesStorage


class TestMoexHistoryJsonParser(unittest.TestCase):

    def setUp(self):
        self.parser = MoexHistoryJsonParser()

        self.expected_close_str = '12345.6789'
        self.expected_result = [SecurityValue(
            trade_date=datetime.date(2000, 12, 31),
            close=decimal.Decimal(self.expected_close_str))]

        self.expected_date_str = self.expected_result[0].trade_date.strftime(self.parser.trade_date_format)

    def test_parse_SuccessWithClose(self):
        valid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE", "SECID", "CLOSE"],
            "data": [["TQBR", "{self.expected_date_str}", "ABRD", {self.expected_close_str}]]
        }}}}"""
        result = list(self.parser.parse(valid_json, tzinfo=None))

        self.assertSequenceEqual(result, self.expected_result)

    def test_parse_AcceptNulls(self):
        expected_result = []
        valid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE", "SECID", "CLOSE"],
            "data": [["TQBR", "{self.expected_date_str}", "ABRD", null]]
        }}}}"""
        result = list(self.parser.parse(valid_json, tzinfo=None))

        self.assertSequenceEqual(result, expected_result)

    def test_parse_SuccessWithLegalClosePrice(self):
        valid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE", "SECID", "LEGALCLOSEPRICE"],
            "data": [["TQBR", "{self.expected_date_str}", "ABRD", {self.expected_close_str}]]
        }}}}"""
        result = list(self.parser.parse(valid_json, tzinfo=None))

        self.assertSequenceEqual(result, self.expected_result)

    def test_parse_SuccessBonds(self):
        face_value = decimal.Decimal(1000)
        self.expected_result[0].close *= face_value / 100
        valid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE", "SECID", "LEGALCLOSEPRICE", "FACEVALUE"],
            "data": [["TQBR", "{self.expected_date_str}", "ABRD", {self.expected_close_str}, {face_value}]]
        }}}}"""
        result = list(self.parser.parse(valid_json, tzinfo=None))

        self.assertSequenceEqual(result, self.expected_result)

    def test_parse_AcceptNullFaceValue(self):
        valid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE", "SECID", "LEGALCLOSEPRICE", "FACEVALUE"],
            "data": [["TQBR", "{self.expected_date_str}", "ABRD", {self.expected_close_str}, null]]
        }}}}"""
        result = list(self.parser.parse(valid_json, tzinfo=None))

        self.assertSequenceEqual(result, self.expected_result)

    def test_parse_raisesWhenNoCloseColumns(self):
        invalid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE", "SECID"],
            "data": [["TQBR", "{self.expected_date_str}", "ABRD"]]
        }}}}"""

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

    def test_parse_raisesWhenWrongDate(self):
        # wrong format
        invalid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE", "SECID"],
            "data": [["TQBR", "9999-9999-9999", "ABRD"]]
        }}}}"""

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

        # wrong type
        invalid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE", "SECID"],
            "data": [["TQBR", 42, "ABRD"]]
        }}}}"""

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

    def test_parse_raisesWhenWrongCloseValue(self):
        # wrong format
        invalid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE", "SECID", "CLOSE"],
            "data": [["TQBR", "{self.expected_date_str}", "ABRD", "123.123.123"]]
        }}}}"""

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

    def test_parse_raisesWhenNoRequiredColumn(self):
        invalid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE__", "SECID", "CLOSE"],
            "data": [["TQBR", "{self.expected_date_str}", "ABRD", {self.expected_close_str}]]
        }}}}"""

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

    def test_parse_raisesWhenNoColumnInData(self):
        invalid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE", "SECID", "CLOSE"],
            "data": [["TQBR", "{self.expected_date_str}", "ABRD"]]
        }}}}"""

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

    def test_parse_raisesWhenNoColumnsBlock(self):
        invalid_json = f"""
        {{
        "history": {{
            "columns__": ["BOARDID", "TRADEDATE", "SECID", "CLOSE"],
            "data": [["TQBR", "{self.expected_date_str}", "ABRD", {self.expected_close_str}]]
        }}}}"""

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

    def test_parse_raisesWhenWrongColumnsType(self):
        invalid_json = f"""
        {{
        "history": {{
            "columns": 42,
            "data": [["TQBR", "{self.expected_date_str}", "ABRD", {self.expected_close_str}]]
        }}}}"""

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

    def test_parse_raisesWhenNoDataBlock(self):
        invalid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE", "SECID", "CLOSE"],
            "data__": [["TQBR", "{self.expected_date_str}", "ABRD", {self.expected_close_str}]]
        }}}}"""

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

    def test_parse_raisesWhenWrongDataType(self):
        invalid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE", "SECID", "CLOSE"],
            "data": 42
        }}}}"""

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

    def test_parse_raisesWhenEmptyHistory(self):
        invalid_json = f"""
        {{
        "history": {{
            "columns": ["BOARDID", "TRADEDATE", "SECID", "CLOSE"],
            "data": []
        }}}}"""

        with self.assertRaises(InstrumentValuesHistoryEmpty):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

    def test_parse_raisesWhenEmptyString(self):
        invalid_json = ''

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

    def test_parse_raisesWhenWrongJson(self):
        invalid_json = """[]"""

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

    def test_parse_raisesWhenUnknownJson(self):
        invalid_json = """{}"""

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))


class TestMoexSecurityInfoJsonParser(unittest.TestCase):

    def setUp(self):
        engine = TradeEngine(identity=42, name='NAME', title='TITLE')
        market = Market(identity=42, trade_engine=engine, name='NAME', title='TITLE', marketplace='MARKETPLACE')
        self.board = Board(
            identity=42,
            trade_engine=engine,
            market=market,
            boardid='BOARDID',
            title='TITLE',
            is_traded=True,
            has_candles=True,
            is_primary=True)
        global_index_data = GlobalIndexData(
            trade_engines=(engine,),
            markets=(market,),
            boards=(self.board,),
        )
        self.parser = MoexSecurityInfoJsonParser(FakeMoexDownloadParameterValuesStorage(global_index_data))

    def test_parse_SuccessAllColumns(self):
        sec_info = SecurityInfo(
            sec_id='ID',
            board=self.board,
            short_name='SHORT NAME',
            lot_size=42,
            sec_name='SEC NAME',
            isin='ISIN',
            lat_name='LAT NAME',
            reg_number='REGNUMBER')
        expected_result = [sec_info]

        json = f"""{{
            "securities": {{
                "columns": ["SECID", "BOARDID", "SHORTNAME", "LOTSIZE", "SECNAME", "NAME", "ISIN", "LATNAME",
                            "REGNUMBER"],
                "data": [["{sec_info.sec_id}", "{sec_info.board.boardid}", "{sec_info.short_name}", 
                           {sec_info.lot_size}, "{sec_info.sec_name}", "{sec_info.sec_name}", "{sec_info.isin}",
                           "{sec_info.lat_name}", "{sec_info.reg_number}"]]
            }}}}"""

        result = list(self.parser.parse(json))

        self.assertSequenceEqual(result, expected_result)

    def test_parse_SuccessOnlyRequiredColumns(self):
        sec_info = SecurityInfo(
            sec_id='ID',
            board=self.board,
            short_name='SHORT NAME')
        expected_result = [sec_info]

        json = f"""{{
            "securities": {{
                "columns": ["SECID", "BOARDID", "SHORTNAME"],
                "data": [["{sec_info.sec_id}", "{sec_info.board.boardid}", "{sec_info.short_name}"]]
            }}}}"""

        result = list(self.parser.parse(json))

        self.assertSequenceEqual(result, expected_result)

    def test_parse_SuccessEmptyList(self):
        expected_result = []

        json = """{
            "securities": {
                "columns": ["SECID", "BOARDID", "SHORTNAME"],
                "data": []
            }}"""

        result = list(self.parser.parse(json))

        self.assertSequenceEqual(result, expected_result)

    def test_parse_raisesWrongBoard(self):
        sec_info = SecurityInfo(
            sec_id='ID',
            board=self.board,
            short_name='SHORT NAME')
        wrong_json = f"""{{
            "securities": {{
                "columns": ["SECID", "BOARDID", "SHORTNAME"],
                "data": [["{sec_info.sec_id}", "WRONG", "{sec_info.short_name}"]]
            }}}}"""

        with self.assertRaises(ParseError):
            list(self.parser.parse(wrong_json))

    def test_parse_raisesEmptyString(self):
        wrong_json = ''

        with self.assertRaises(ParseError):
            list(self.parser.parse(wrong_json))

    def test_parse_raisesWrongJson(self):
        wrong_json = '[]'

        with self.assertRaises(ParseError):
            list(self.parser.parse(wrong_json))


class TestMoexGlobalIndexJsonParser(unittest.TestCase):

    def setUp(self):
        self.engine = TradeEngine(
            identity=1,
            name='stock',
            title='Фондовый рынок и рынок депозитов')
        self.market = Market(
            identity=1,
            trade_engine=self.engine,
            name='shares',
            title='Рынок акций',
            marketplace='MXSE')
        self.board = Board(
            identity=178,
            trade_engine=self.engine,
            market=self.market,
            boardid='TQTF',
            title='Т+: ETF - безадрес.',
            is_traded=True,
            has_candles=True,
            is_primary=True)
        self.expected_result = GlobalIndexData(
            trade_engines=(self.engine,),
            markets=(self.market,),
            boards=(self.board,))

        self.parser = MoexGlobalIndexJsonParser()

    def generate_valid_json(self):
        valid_json = f"""
        {{
        "engines": {{
            "metadata": {{
                "id": {{"type": "int32"}},
                "name": {{"type": "string", "bytes": 45, "max_size": 0}},
                "title": {{"type": "string", "bytes": 765, "max_size": 0}}
            }},
            "columns": ["id", "name", "title"], 
            "data": [[{self.engine.identity}, "{self.engine.name}", "{self.engine.title}"]]
        }},
        "markets": {{
            "metadata": {{
                "id": {{"type": "int32"}},
                "trade_engine_id": {{"type": "int32"}},
                "trade_engine_name": {{"type": "string", "bytes": 45, "max_size": 0}},
                "trade_engine_title": {{"type": "string", "bytes": 765, "max_size": 0}},
                "market_name": {{"type": "string", "bytes": 45, "max_size": 0}},
                "market_title": {{"type": "string", "bytes": 765, "max_size": 0}},
                "market_id": {{"type": "int32"}},
                "marketplace": {{"type": "string", "bytes": 48, "max_size": 0}}
            }},
            "columns": ["id", "trade_engine_id", "trade_engine_name", "trade_engine_title", "market_name", 
                        "market_title", "market_id", "marketplace"], 
            "data": [[{self.market.identity}, {self.engine.identity}, "{self.engine.name}", "{self.engine.title}",
                      "{self.market.name}", "{self.market.title}", {self.market.identity}, "{self.market.marketplace}"]]
        }},
        "boards": {{
            "metadata": {{
                "id": {{"type": "int32"}},
                "board_group_id": {{"type": "int32"}},
                "engine_id": {{"type": "int32"}},
                "market_id": {{"type": "int32"}},
                "boardid": {{"type": "string", "bytes": 12, "max_size": 0}},
                "board_title": {{"type": "string", "bytes": 381, "max_size": 0}},
                "is_traded": {{"type": "int32"}},
                "has_candles": {{"type": "int32"}},
                "is_primary": {{"type": "int32"}}
            }},
            "columns": ["id", "board_group_id", "engine_id", "market_id", "boardid", "board_title", "is_traded",
                        "has_candles", "is_primary"], 
            "data": [[{self.board.identity}, 57, {self.engine.identity}, {self.market.identity},
                      "{self.board.boardid}", "{self.board.title}",
                      {1 if self.board.is_traded else 0}, {1 if self.board.has_candles else 0}, 
                      {1 if self.board.is_primary else 0}]]
        }}
        }}"""

        return valid_json

    def spoil_block_data(self, block_name: str, new_data_content: str):
        spoil_pattern = re.compile(
            fr'"{block_name}"\s*:\s*{{(.*?)"data"\s*:\s*.*?}}',
            re.IGNORECASE | re.MULTILINE | re.DOTALL)
        return spoil_pattern.sub(fr'"{block_name}":{{\1"data":{new_data_content}}}', self.generate_valid_json())

    def test_parse_Success(self):
        result = self.parser.parse(self.generate_valid_json())

        self.assertEqual(result, self.expected_result)

    def test_parse_raisesEmptyString(self):
        wrong_json = ''

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(wrong_json))

    def test_parse_raisesWrongJson(self):
        wrong_json = '[]'

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(wrong_json))

    def test_parse_raisesWhenWrongEnginesContent(self):
        # no block
        wrong_json = self.generate_valid_json().replace(f'"engines"', f'"engines__"')

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(wrong_json))

        # bad data content type
        wrong_json = self.spoil_block_data('engines', '{}')

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(wrong_json))

    def test_parse_raisesWhenWrongMarketsContent(self):
        # no block
        wrong_json = self.generate_valid_json().replace(f'"markets"', f'"markets__"')

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(wrong_json))

        # bad data content type
        wrong_json = self.spoil_block_data('markets', '{}')

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(wrong_json))

        # bad engine id
        wrong_json = self.spoil_block_data(
            'markets',
            f"""[[{self.market.identity}, -1, "{self.engine.name}", "{self.engine.title}",
                  "{self.market.name}", "{self.market.title}", {self.market.identity}, "{self.market.marketplace}"]]""")

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(wrong_json))

    def test_parse_raisesWhenWrongBoardsContent(self):
        # no block
        wrong_json = self.generate_valid_json().replace(f'"boards"', f'"boards__"')

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(wrong_json))

        # bad data content type
        wrong_json = self.spoil_block_data('boards', '{}')

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(wrong_json))

        # bad engine id
        wrong_json = self.spoil_block_data(
            'boards',
            f"""[[{self.board.identity}, 57, -1, {self.market.identity},
                  "{self.board.boardid}", "{self.board.title}",
                  {1 if self.board.is_traded else 0}, {1 if self.board.has_candles else 0}, 
                  {1 if self.board.is_primary else 0}]]""")

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(wrong_json))

        # bad market id
        wrong_json = self.spoil_block_data(
            'boards',
            f"""[[{self.board.identity}, 57, {self.engine.identity}, -1,
                  "{self.board.boardid}", "{self.board.title}",
                  {1 if self.board.is_traded else 0}, {1 if self.board.has_candles else 0}, 
                  {1 if self.board.is_primary else 0}]]""")

        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(wrong_json))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import typing
import datetime

from sane_finances.communication.downloader import Downloader, DownloadStringResult
from sane_finances.communication.cachers import ExpiryCalculator, DummyCacher, InMemoryCacher


class FakeDummyCacher(DummyCacher):

    def __init__(self):
        self.retrieve_count = 0
        self.drop_count = 0
        self.has_count = 0
        self.is_empty_count = 0
        self.full_clear_count = 0

    def retrieve(
            self,
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str],
            reviver: typing.Callable[[], str]) -> typing.Tuple[bool, str]:
        self.retrieve_count += 1
        return super().retrieve(url, parameters, headers, reviver)

    def drop(
            self,
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str]) -> bool:
        self.drop_count += 1
        return super().drop(url, parameters, headers)

    def has(
            self,
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str]) -> bool:
        self.has_count += 1
        return super().has(url, parameters, headers)

    def is_empty(self) -> bool:
        """ True if cache is empty
        """
        self.is_empty_count += 1
        return super().is_empty()

    def full_clear(self):
        """ Full clear of internal storage.
        """
        self.full_clear_count += 1
        return super().full_clear()


class FakeExpirableCacher(InMemoryCacher):
    # create full fake of 'ExpirableCacher' is very time-consuming and laborious
    # so assume that 'InMemoryCacher' is absolutely correct (:wink:) and take it
    pass


class FakeExpiryCalculator(ExpiryCalculator):

    def __init__(self, now: datetime.datetime):
        self.now = now

    def is_expired(self, expiry_moment: datetime.datetime) -> bool:
        return self.now > expiry_moment

    def get_expiry_moment(self, delta: datetime.timedelta, start_from: datetime.datetime = None) -> datetime.datetime:
        if start_from is None:
            start_from = self.now
        return start_from + delta

    def get_revive_moment(self) -> datetime.datetime:
        return self.now


class FakeDownloader(Downloader):

    def __init__(self, fake_data: typing.Optional[str]):
        self.fake_data = fake_data
        self.download_string_results: typing.List[DownloadStringResult] = []

    @property
    def parameters(self) -> typing.List[typing.Tuple[str, str]]:
        return []

    @parameters.setter
    def parameters(self, value) -> None:
        pass

    @property
    def headers(self) -> typing.Dict[str, str]:
        return {}

    @headers.setter
    def headers(self, value) -> None:
        pass

    def download_string(self, *args, **kwargs) -> DownloadStringResult:
        result = DownloadStringResult(self.fake_data)
        self.download_string_results.append(result)
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import typing
import datetime

from sane_finances.communication.downloader import DownloadStringResult
from sane_finances.sources.msci.v2021.meta import (
    IndexValue, IndexInfo, IndexPanelData)
from sane_finances.sources.msci.v2021.parsers import (
    MsciHistoryJsonParser, MsciIndexInfoParser, MsciIndexPanelDataJsonParser)
from sane_finances.sources.msci.v2021.exporters import (
    MsciStringDataDownloader, MsciIndexDownloadParameterValuesStorage)

from ....communication.fakes import FakeDownloader


class FakeMsciIndexPanelDataJsonParser(MsciIndexPanelDataJsonParser):

    def __init__(self, fake_index_panel_data: IndexPanelData):
        super().__init__()

        self.fake_index_panel_data = fake_index_panel_data
        self.parse_exception = None

    def parse(self, raw_json_text: str) -> IndexPanelData:
        if self.parse_exception is not None:
            raise self.parse_exception

        return self.fake_index_panel_data


class FakeMsciIndexDownloadParameterValuesStorage(MsciIndexDownloadParameterValuesStorage):

    def __init__(self, fake_index_panel_data: IndexPanelData):
        super().__init__(FakeDownloader(None), FakeMsciIndexPanelDataJsonParser(fake_index_panel_data))


class FakeIndexInfoParser(MsciIndexInfoParser):

    def __init__(self, fake_data: typing.Iterable[IndexInfo]):
        super().__init__()

        self.fake_data = fake_data
        self.parse_exception = None

    def parse(self, xml) -> typing.Iterable[IndexInfo]:
        if self.parse_exception is not None:
            raise self.parse_exception

        return self.fake_data


class FakeMsciHistoryJsonParser(MsciHistoryJsonParser):

    def __init__(self, fake_data: typing.Iterable[IndexValue]):
        # noinspection PyTypeChecker
        super().__init__(None)

        self.fake_data = fake_data
        self.parse_exception = None

    def parse(
            self,
            raw_xml_text: str,
            tzinfo: typing.Optional[datetime.timezone]
    ) -> typing.Iterable[IndexValue]:
        if self.parse_exception is not None:
            raise self.parse_exception

        return self.fake_data


class FakeMsciStringDataDownloader(MsciStringDataDownloader):

    def __init__(self, fake_info_data, fake_history_data):
        # noinspection PyTypeChecker
        super().__init__(FakeDownloader(None), None)

        self.download_instruments_info_string_results: typing.List[DownloadStringResult] = []
        self.download_instrument_history_string_results: typing.List[DownloadStringResult] = []

        self.fake_history_data = fake_history_data
        self.fake_info_data = fake_info_data

    def download_index_history_string(
            self,
            index_code,
            currency,
            index_variant,
            date_from,
            date_to) -> DownloadStringResult:
        result = DownloadStringResult(self.fake_history_data)
        self.download_instrument_history_string_results.append(result)
        return result

    def download_indexes_info_string(
            self,
            scope,
            market,
            size,
            style,
            index_suite) -> DownloadStringResult:
        result = DownloadStringResult(self.fake_info_data)
        self.download_instruments_info_string_results.append(result)
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import decimal
import unittest

from sane_finances.sources.base import (
    InstrumentValue, InstrumentInfo, DownloadParametersFactory, InstrumentHistoryDownloadParameters)
from sane_finances.sources.cbr.v2016.meta import (
    RateFrequencies, CurrencyInfo, CurrencyRateValue,
    CbrCurrenciesInfoDownloadParameters, CbrCurrencyHistoryDownloadParameters, CbrDownloadParametersFactory)
from .common import CommonTestCases


class TestCurrencyRateValue(unittest.TestCase):

    def test_instrument_value_Success(self):
        index_value = CurrencyRateValue(
            date=datetime.date(2000, 12, 31),
            value=decimal.Decimal(42),
            nominal=42,
            currency_id='ID')
        expected_instrument_value = InstrumentValue(
            value=index_value.value / index_value.nominal,
            moment=datetime.datetime.combine(index_value.date, datetime.time.min, tzinfo=datetime.timezone.utc))

        instrument_value = index_value.get_instrument_value(datetime.timezone.utc)

        self.assertEqual(expected_instrument_value, instrument_value)

    def test_raiseWrongDate(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = CurrencyRateValue(
                date=None,
                value=decimal.Decimal(42),
                nominal=42,
                currency_id='ID')


class TestCurrencyInfo(unittest.TestCase):

    def test_instrument_value_Success(self):
        index_info = CurrencyInfo(
            currency_id='ID',
            name='NAME',
            eng_name='ENG NAME',
            nominal=42,
            parent_code='PARENT CODE')
        expected_instrument_info = InstrumentInfo(code=index_info.currency_id, name=index_info.name)

        instrument_info = index_info.instrument_info

        self.assertEqual(expected_instrument_info, instrument_info)


class TestCbrCurrenciesInfoDownloadParameters(unittest.TestCase):

    def test_safe_create_Success(self):
        _ = CbrCurrenciesInfoDownloadParameters.safe_create(rate_frequency=RateFrequencies.DAILY)

    def test_safe_create_RaiseWrongFrequency(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = CbrCurrenciesInfoDownloadParameters.safe_create(rate_frequency=None)


class TestCbrCurrencyHistoryDownloadParameters(unittest.TestCase):

    def test_safe_create_Success(self):
        _ = CbrCurrencyHistoryDownloadParameters.safe_create(currency_id='ID')

    def test_clone_with_instrument_info_parameters_Success(self):
        params = CbrCurrencyHistoryDownloadParameters.safe_create(currency_id='ID')
        expected_result = params

        result = params.clone_with_instrument_info_parameters(None, None)

        self.assertEqual(expected_result, result)


class TestCbrDownloadParametersFactory(CommonTestCases.CommonDownloadParametersFactoryTests):

    def get_download_parameters_factory(self) -> DownloadParametersFactory:
        return CbrDownloadParametersFactory()

    def get_generate_history_download_parameters_from_expected_result_with_none(self) \
            -> InstrumentHistoryDownloadParameters:
        # noinspection PyTypeChecker
        expected_result = CbrCurrencyHistoryDownloadParameters(currency_id=None)
        return expected_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Tools for download (export) index data from moex.com
"""
import collections
import dataclasses
import decimal
import itertools
import logging
import typing
import datetime
import inspect

from .meta import (
    TradeEngine, Market, Board, ResponseFormats, JsonFormats, Limits,
    GlobalIndexData,
    MoexSecurityHistoryDownloadParameters, MoexSecuritiesInfoDownloadParameters,
    MoexDownloadParametersFactory)
from .parsers import MoexGlobalIndexJsonParser, MoexHistoryJsonParser, MoexSecurityInfoJsonParser
from ...base import (
    ApiActualityChecker, InstrumentStringDataDownloader, ParseError,
    InstrumentExporterFactory, InstrumentHistoryValuesExporter, InstrumentsInfoExporter,
    DownloadParameterValuesStorage, DynamicEnumTypeManager, DownloadStringResult,
    SourceError, CheckApiActualityError)
from ...generic import GenericInstrumentHistoryValuesExporter, GenericInstrumentsInfoExporter
from ....communication.downloader import Downloader

logging.getLogger().addHandler(logging.NullHandler())


class MoexStringDataDownloader(InstrumentStringDataDownloader):
    """ Data downloader from moex.com.
    """

    BaseUrl = 'https://iss.moex.com/iss'

    def __init__(self, downloader: Downloader):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.downloader = downloader

        # headers for HTTP
        self.headers: typing.Dict[str, str] = {}

        self.iss_meta = False
        self.iss_data = True
        self.iss_json = JsonFormats.COMPACT
        self.limit = Limits.HUNDRED
        self.response_format = ResponseFormats.JSON
        self.history_columns = ['BOARDID', 'TRADEDATE', 'SECID', 'CLOSE', 'LEGALCLOSEPRICE', 'FACEVALUE']

    def paginate_download_instrument_history_parameters(
            self,
            parameters: MoexSecurityHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime
    ) -> typing.Iterable[typing.Tuple[MoexSecurityHistoryDownloadParameters, datetime.datetime, datetime.datetime]]:

        for start in itertools.count(start=0, step=self.limit.value):
            parameters = dataclasses.replace(parameters, start=start)
            yield parameters, moment_from, moment_to

    def download_instrument_history_string(
            self,
            parameters: MoexSecurityHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime) -> DownloadStringResult:
        date_from = moment_from.date()
        date_to = moment_to.date()
        if moment_to.time() != datetime.time.min:
            date_to += datetime.timedelta(days=1)

        return self.download_security_history_string(
            parameters.board,
            parameters.sec_id,
            parameters.start,
            date_from,
            date_to)

    def download_instruments_info_string(
            self,
            parameters: MoexSecuritiesInfoDownloadParameters) -> DownloadStringResult:
        return self.download_securities_info_string(
            parameters.board)

    def download_security_history_string(
            self,
            board: Board,
            sec_id: str,
            start: int,
            date_from: datetime.date,
            date_to: datetime.date) -> DownloadStringResult:
        """ Downloads history data for one security as string.

        :param board: Board.
        :param sec_id: Security ID.
        :param start: Start value.
        :param date_from: Download interval beginning.
        :param date_to: Download interval ending.
        :return: Container with downloaded string.
        """
        params = [
            ('iss.meta', 'on' if self.iss_meta else 'off'),
            ('iss.data', 'on' if self.iss_data else 'off'),
            ('iss.json', str(self.iss_json.value)),
            ('limit', str(self.limit.value)),
            ('start', str(start)),
            ('history.columns', ','.join(self.history_columns)),
            ('from', date_from.strftime('%Y-%m-%d')),
            ('till', date_to.strftime('%Y-%m-%d'))
        ]

        url = f"{self.BaseUrl}/history/engines/{board.trade_engine.name}/" \
              f"markets/{board.market.name}/" \
              f"boards/{board.boardid}/" \
              f"securities/{sec_id}.{self.response_format.value}"

        self.downloader.parameters = params
        self.downloader.headers = self.headers

        return self.downloader.download_string(url)

    def download_securities_info_string(
            self,
            board: Board) -> DownloadStringResult:
        """ Downloads the list of all available securities by specified parameters.

        :param board: Board.
        :return: Container with downloaded string.
        """
        self.downloader.headers = self.headers
        self.downloader.parameters = []
        url = f"{self.BaseUrl}/engines/{board.trade_engine.name}/" \
              f"markets/{board.market.name}/" \
              f"boards/{board.boardid}/" \
              f"securities.{self.response_format.value}"

        return self.downloader.download_string(url)


class MoexDynamicEnumTypeManager(DynamicEnumTypeManager):
    """ MOEX dynamic enum types manager.
    """

    def __init__(self):
        self._managed_types: typing.Dict[typing.Type, typing.Tuple[str, typing.Callable]] = {
            TradeEngine: ('trade_engines', lambda it: it.identity),
            Market: ('markets', lambda it: it.identity),
            Board: ('boards', lambda it: it.identity)
        }

    def is_dynamic_enum_type(self, cls: type) -> bool:
        if not inspect.isclass(cls):
            return False

        return cls in self._managed_types

    def get_all_managed_types(self) -> typing.Iterable[typing.Type]:
        return self._managed_types.keys()

    def get_dynamic_enum_key(self, instance):
        for managed_type, (_, key_getter, *_) in self._managed_types.items():
            if isinstance(instance, managed_type):
                return key_getter(instance)

        return None


class MoexDownloadParameterValuesStorage(MoexDynamicEnumTypeManager, DownloadParameterValuesStorage):
    """ Storage of instrument download parameters.
    """

    def __init__(
            self,
            downloader: Downloader,
            global_index_json_parser: MoexGlobalIndexJsonParser):
        super().__init__()
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.downloader = downloader
        self.global_index_json_parser = global_index_json_parser

        # headers for HTTP
        self.headers: typing.Dict[str, str] = {}

        self._extended_managed_types: typing.Dict[typing.Type, typing.Tuple] = {
            TradeEngine: (self._get_trade_engines_choices,),
            Market: (self._get_markets_choices,),
            Board: (self._get_boards_choices,)
        }
        assert set(self._extended_managed_types.keys()) == set(self._managed_types.keys()), \
            ("MoexDynamicEnumTypeManager and MoexDownloadParameterValuesStorage "
             "has different managed types.")

        self.global_index_data: typing.Optional[GlobalIndexData] = None

    def reload(self) -> None:
        self.downloader.headers = self.headers
        self.downloader.parameters = []
        url = f"{MoexStringDataDownloader.BaseUrl}/index.json"

        json_string_result = self.downloader.download_string(url)

        try:
            global_index_data = self.global_index_json_parser.parse(json_string_result.downloaded_string)

        except Exception:
            json_string_result.set_correctness(False)
            raise

        json_string_result.set_correctness(True)
        self.global_index_data = global_index_data

    def _ensure_loaded(self):
        if self.global_index_data is None:
            self.reload()

    def get_dynamic_enum_value_by_key(self, cls: type, key) -> typing.Any:
        if not self.is_dynamic_enum_type(cls):
            return None

        self._ensure_loaded()

        global_index_data_attr_name, key_getter, *_ = self._managed_types[cls]
        enum_values = getattr(self.global_index_data, global_index_data_attr_name)

        for enum_value in enum_values:
            if key == key_getter(enum_value):
                return enum_value

        return None

    def get_dynamic_enum_value_by_choice(self, cls: type, choice) -> typing.Any:
        if not self.is_dynamic_enum_type(cls):
            return None

        self._ensure_loaded()

        try:
            # here we know that all choices from get_parameter_type_choices
            # was made by converting identities to strings,
            # so we convert them back
            key = int(choice)
        except (ValueError, TypeError) as ex:
            raise SourceError(f"Can't get enum value key from {choice!r}") from ex

        global_index_data_attr_name, key_getter, *_ = self._managed_types[cls]
        enum_values = getattr(self.global_index_data, global_index_data_attr_name)

        for enum_value in enum_values:
            if key == key_getter(enum_value):
                return enum_value

        return None

    def get_all_parameter_values_for(self, cls: type) -> typing.Optional[typing.Iterable]:
        if not self.is_dynamic_enum_type(cls):
            return None

        self._ensure_loaded()

        global_index_data_attr_name, *_ = self._managed_types[cls]
        return getattr(self.global_index_data, global_index_data_attr_name)

    def get_parameter_type_choices(self, cls: type) \
            -> typing.Optional[
                typing.List[typing.Tuple[typing.Any, typing.Union[str, typing.List[typing.Tuple[typing.Any, str]]]]]
            ]:
        if not self.is_dynamic_enum_type(cls):
            return None

        self._ensure_loaded()

        choices_getter, *_ = self._extended_managed_types[cls]
        return choices_getter()

    def _get_trade_engines_choices(self):
        return [(str(trade_engine.identity), trade_engine.title)
                for trade_engine
                in self.global_index_data.trade_engines]

    def _get_markets_choices(self):
        market: Market

        grouped_markets = collections.OrderedDict()
        for market in self.global_index_data.markets:
            if market.trade_engine not in grouped_markets:
                grouped_markets[market.trade_engine] = []
            grouped_markets[market.trade_engine].append(market)

        return [(f"Available for trade engine {trade_engine.title!r}:",
                 [(str(market.identity), market.title) for market in markets])
                for trade_engine, markets
                in grouped_markets.items()]

    def _get_boards_choices(self):
        board: Board

        grouped_boards = collections.OrderedDict()
        for board in self.global_index_data.boards:
            key = (board.trade_engine, board.market)
            if key not in grouped_boards:
                grouped_boards[key] = []
            grouped_boards[key].append(board)

        return [(f"Available for {trade_engine.title!r}, {market.title!r}:",
                 [(str(board.identity), board.title) for board in boards])
                for (trade_engine, market), boards
                in grouped_boards.items()]


class MoexApiActualityChecker(ApiActualityChecker):
    """ Verifies actuality and accessibility of REST API of moex.com.
    """

    _trade_engine_name_to_test = 'stock'
    _market_name_to_test = 'index'
    _boardid_to_test = 'TQTF'
    _sec_id_to_test = 'FXJP'
    _history_date_to_test = datetime.date(2014, 6, 10)
    _expected_close_value = decimal.Decimal('1000')

    def __init__(
            self,
            parameter_values_storage: MoexDownloadParameterValuesStorage,
            string_data_downloader: MoexStringDataDownloader,
            info_parser: MoexSecurityInfoJsonParser,
            history_values_parser: MoexHistoryJsonParser):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.parameter_values_storage = parameter_values_storage
        self.string_data_downloader = string_data_downloader
        self.history_values_parser = history_values_parser
        self.info_parser = info_parser

    def check(self):
        self.logger.info("Check actuality via security list")

        all_boards: typing.Tuple[Board] = tuple(self.parameter_values_storage.get_all_parameter_values_for(Board))
        if not all_boards:
            raise CheckApiActualityError("Not found any board")

        # try to find any board for stock indexes
        all_stock_index_boards = [board
                                  for board
                                  in all_boards
                                  if (board.trade_engine.name == self._trade_engine_name_to_test
                                      and board.market.name == self._market_name_to_test)]

        if not all_stock_index_boards:
            # else get any other board
            all_stock_index_boards = [all_boards[0]]

        target_board = all_stock_index_boards[0]
        securities_info_string_result = self.string_data_downloader.download_securities_info_string(target_board)
        self.logger.debug(f"Got JSON data:\n{securities_info_string_result.downloaded_string}")
        # read all available securities
        try:
            _ = tuple(self.info_parser.parse(securities_info_string_result.downloaded_string))
        except ParseError as ex:
            securities_info_string_result.set_correctness(False)
            raise CheckApiActualityError(f"Unexpected securities info JSON: {ex.message}") from ex
        except Exception:
            securities_info_string_result.set_correctness(False)
            raise

        # now test history data
        self.logger.info(f"Check actuality via security {self._sec_id_to_test!r}")

        target_board = ([board
                         for board
                         in all_boards
                         if board.boardid == self._boardid_to_test] or [None])[0]

        if target_board is None:
            securities_info_string_result.set_correctness(False)
            raise CheckApiActualityError(f"Not found board with boardid {self._boardid_to_test!r}")

        history_data_string_result = self.string_data_downloader.download_security_history_string(
            target_board,
            self._sec_id_to_test,
            0,
            self._history_date_to_test,
            self._history_date_to_test
        )
        self.logger.debug(f"Got JSON data:\n{history_data_string_result.downloaded_string}")
        try:
            history_data = tuple(self.history_values_parser.parse(
                history_data_string_result.downloaded_string,
                tzinfo=None))
        except ParseError as ex:
            history_data_string_result.set_correctness(False)
            raise CheckApiActualityError(f"Unexpected securities history JSON: {ex.message}") from ex
        except Exception:
            history_data_string_result.set_correctness(False)
            raise

        if not history_data:
            history_data_string_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Not found history values for {self._sec_id_to_test!r} "
                f"on {self._history_date_to_test.isoformat()}")

        value_to_test = history_data[0]

        if value_to_test.trade_date != self._history_date_to_test:
            history_data_string_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Unexpected trade date. {value_to_test.trade_date!r} != {self._history_date_to_test!r}")

        if value_to_test.close != self._expected_close_value:
            history_data_string_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Unexpected close value. {value_to_test.close!r} != {self._expected_close_value!r}")

        self.logger.info("Actuality check was successful")


# noinspection PyPep8Naming
class MoexIndexExporterFactory_v1_3(InstrumentExporterFactory):  # pylint: disable=invalid-name
    """ Factory class for create instances of Moscow Exchange data exporter.
    """
    name: str = 'Moscow Exchange ISS data exporter. ' \
                'X-Micex-ISS-Query-Version: 1.0. ' \
                'X-Micex-ISS-Statement-Version: history=3.0'
    provider_site: str = 'https://www.moex.com/'
    api_url: str = 'http://iss.moex.com/iss/reference/'

    def __init__(self):
        self._dynamic_enum_type_manager = None
        self._download_parameters_factory = None

    def create_history_values_exporter(self, downloader: Downloader) -> InstrumentHistoryValuesExporter:
        string_data_downloader = MoexStringDataDownloader(downloader)
        history_values_parser = MoexHistoryJsonParser()

        return GenericInstrumentHistoryValuesExporter(string_data_downloader, history_values_parser)

    def create_info_exporter(self, downloader: Downloader) -> InstrumentsInfoExporter:
        string_data_downloader = MoexStringDataDownloader(downloader)
        global_index_json_parser = MoexGlobalIndexJsonParser()
        parameter_values_storage = MoexDownloadParameterValuesStorage(downloader, global_index_json_parser)
        info_parser = MoexSecurityInfoJsonParser(parameter_values_storage)

        return GenericInstrumentsInfoExporter(string_data_downloader, info_parser)

    def create_download_parameter_values_storage(self, downloader: Downloader) -> MoexDownloadParameterValuesStorage:
        global_index_json_parser = MoexGlobalIndexJsonParser()

        return MoexDownloadParameterValuesStorage(downloader, global_index_json_parser)

    def create_api_actuality_checker(self, downloader: Downloader) -> MoexApiActualityChecker:
        string_data_downloader = MoexStringDataDownloader(downloader)
        history_values_parser = MoexHistoryJsonParser()
        global_index_json_parser = MoexGlobalIndexJsonParser()
        parameter_values_storage = MoexDownloadParameterValuesStorage(downloader, global_index_json_parser)
        info_parser = MoexSecurityInfoJsonParser(parameter_values_storage)

        return MoexApiActualityChecker(
            parameter_values_storage,
            string_data_downloader,
            info_parser,
            history_values_parser)

    @property
    def dynamic_enum_type_manager(self) -> MoexDynamicEnumTypeManager:
        if self._dynamic_enum_type_manager is None:
            self._dynamic_enum_type_manager = MoexDynamicEnumTypeManager()

        return self._dynamic_enum_type_manager

    @property
    def download_parameters_factory(self) -> MoexDownloadParametersFactory:
        if self._download_parameters_factory is None:
            self._download_parameters_factory = MoexDownloadParametersFactory()

        return self._download_parameters_factory



#!/usr/bin/python
# -*- coding: utf-8 -*-

import unittest

from sane_finances.sources.inspection import InstrumentInfoParameter


class TestInstrumentInfoParameter(unittest.TestCase):

    def test_Success(self):
        _ = InstrumentInfoParameter()
        _ = InstrumentInfoParameter(instrument_identity=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Metadata for export index data from app2.msci.com
"""

import decimal
import typing
import datetime
import enum
import dataclasses

from ...base import (
    InstrumentValue, InstrumentInfo, InstrumentValueProvider, InstrumentInfoProvider,
    InstrumentHistoryDownloadParameters, DownloadParametersFactory)
from ...inspection import InstrumentInfoParameter
from ....annotations import LEGACY_ANNOTATIONS

if LEGACY_ANNOTATIONS:  # pragma: no cover
    from ....annotations import Annotated
else:  # pragma: no cover
    from typing import Annotated


class Scopes(enum.Enum):
    """ Index scope.
    """

    def __new__(cls, value: str, description: str):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = str(description)
        return obj

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}: " \
               f"'{self.value}' ('{self.description}')>"

    REGIONAL = ('Region', 'Regional')
    COUNTRY = ('Country', 'Country')


class Market(typing.NamedTuple):
    """ Market from msci.com.
    """
    identity: str
    name: str
    scope: typing.Optional[Scopes] = None

    @classmethod
    def safe_create(
            cls: typing.Type['Market'],
            *,
            identity: str,
            name: str,
            scope: Scopes = None) -> 'Market':
        """ Create new instance of ``Market`` with arguments check.

        :param identity: Identity value.
        :param name: Name.
        :param scope: Scope. ``None`` if market is available for all scopes.
        :return: ``Market`` instance.
        """
        if scope is not None and not isinstance(scope, Scopes):
            raise TypeError("'scope' is not Scopes")

        # see https://github.com/PyCQA/pylint/issues/1801 for pylint disable hint details
        return cls(
            identity=str(identity),
            name=str(name),
            scope=None if scope is None else Scopes(scope))  # pylint: disable=no-value-for-parameter


class Currency(typing.NamedTuple):
    """ Index currency.

    The MSCI Country and Regional Indices are calculated in local currency as well as in USD.
    The concept of a “local currency” calculation excludes the impact of currency fluctuations.
    """
    identity: str
    name: str

    @classmethod
    def safe_create(
            cls: typing.Type['Currency'],
            *,
            identity: str,
            name: str) -> 'Currency':
        """ Create new instance of ``Currency`` with arguments check.

        :param identity: Identity value.
        :param name: Currency name.
        :return: ``Currency`` instance.
        """
        return cls(identity=str(identity), name=str(name))


class IndexLevel(typing.NamedTuple):
    """ Index level (or price level).

    Total return indices measure the market performance,
    including price performance and income from regular cash distributions
    (cash dividend payments or capital repayments).

    Gross Daily Total Return: This series approximates the maximum possible reinvestment
    of regular cash distributions (dividends or capital repayments).
    The amount reinvested is the cash distributed to individuals resident in the country of the company,
    but does not include tax credits.

    Net Daily Total Return: This series approximates the minimum possible reinvestment
    of regular cash distributions. Provided that the regular capital repayment is not subject to withholding tax,
    the reinvestment in the Net Daily Total Return is free of withholding tax.

    The Total Return Index that represents the weighted return of the MSCI parent index
    and the cash component.

    The Excess Return Index that represents the return of the Total Return Index
    minus the return of the cash component.
    """
    identity: str
    name: str

    @classmethod
    def safe_create(
            cls: typing.Type['IndexLevel'],
            *,
            identity: str,
            name: str) -> 'IndexLevel':
        """ Create new instance of ``IndexLevel`` with arguments check.

        :param identity: Identity value.
        :param name: Index level name.
        :return: ``IndexLevel`` instance.
        """
        return cls(identity=str(identity), name=str(name))


class Frequency(typing.NamedTuple):
    """ Index values frequency.
    """
    identity: str
    name: str

    @classmethod
    def safe_create(
            cls: typing.Type['Frequency'],
            *,
            identity: str,
            name: str) -> 'Frequency':
        """ Create new instance of ``Frequency`` with arguments check.

        :param identity: Identity value.
        :param name: Frequency name.
        :return: ``Frequency`` instance.
        """
        return cls(identity=str(identity), name=str(name))


class Style(typing.NamedTuple):
    """ Index style.
    """
    identity: str
    name: str

    @classmethod
    def safe_create(
            cls: typing.Type['Style'],
            *,
            identity: str,
            name: str) -> 'Style':
        """ Create new instance of ``Style`` with arguments check.

        :param identity: Identity value.
        :param name: Style name.
        :return: ``Style`` instance.
        """
        return cls(identity=str(identity), name=str(name))


class Size(typing.NamedTuple):
    """ Index size.
    """
    identity: str
    name: str

    @classmethod
    def safe_create(
            cls: typing.Type['Size'],
            *,
            identity: str,
            name: str) -> 'Size':
        """ Create new instance of ``Size`` with arguments check.

        :param identity: Identity value.
        :param name: Size name.
        :return: ``Size`` instance.
        """
        return cls(identity=str(identity), name=str(name))


class IndexSuiteGroup(typing.NamedTuple):
    """ Group of index suites.
    """
    name: str

    @classmethod
    def safe_create(
            cls: typing.Type['IndexSuiteGroup'],
            *,
            name: str) -> 'IndexSuiteGroup':
        """ Create new instance of ``IndexSuiteGroup`` with arguments check.

        :param name: Index suite group name.
        :return: ``IndexSuiteGroup`` instance.
        """
        return cls(name=str(name))


class IndexSuite(typing.NamedTuple):
    """ Index suite.
    """
    identity: str
    name: str
    group: typing.Optional[IndexSuiteGroup] = None

    @classmethod
    def safe_create(
            cls: typing.Type['IndexSuite'],
            *,
            identity: str,
            name: str,
            group: typing.Optional[IndexSuiteGroup] = None) -> 'IndexSuite':
        """ Create new instance of ``IndexSuite`` with arguments check.

        :param identity: Identity value.
        :param name: Index suite name.
        :param group: Index suite group. ``None`` if index suite is not inside any group.
        :return: ``IndexSuite`` instance.
        """
        if group is not None and not isinstance(group, IndexSuiteGroup):
            raise TypeError("'group' is not IndexSuiteGroup")

        return cls(identity=str(identity), name=str(name), group=group)


class IndexPanelData(typing.NamedTuple):
    """ Container for index panel data from msci.com.
    """
    markets: typing.Tuple[Market, ...]
    currencies: typing.Tuple[Currency, ...]
    index_levels: typing.Tuple[IndexLevel, ...]
    frequencies: typing.Tuple[Frequency, ...]
    index_suite_groups: typing.Tuple[IndexSuiteGroup, ...]
    index_suites: typing.Tuple[IndexSuite, ...]
    sizes: typing.Tuple[Size, ...]
    styles: typing.Tuple[Style, ...]
    daily_frequency: Frequency
    monthly_frequency: Frequency


@dataclasses.dataclass
class IndexValue(InstrumentValueProvider):
    """ Container for index history item.
    """
    calc_date: datetime.date
    level_eod: decimal.Decimal
    msci_index_code: str
    index_variant_type: IndexLevel
    currency: Currency

    def __init__(self,
                 *,
                 calc_date: datetime.date,
                 level_eod: decimal.Decimal,
                 msci_index_code: str,
                 index_variant_type: IndexLevel,
                 currency: Currency):
        if not isinstance(calc_date, datetime.date):
            raise TypeError("'calc_date' is not date")
        if not isinstance(index_variant_type, IndexLevel):
            raise TypeError("'index_variant_type' is not IndexLevel")
        if not isinstance(currency, Currency):
            raise TypeError("'currency' is not Currency")

        self.calc_date = calc_date
        self.level_eod = decimal.Decimal(level_eod)

        self.msci_index_code = str(msci_index_code)
        self.index_variant_type = index_variant_type
        self.currency = currency

    def get_instrument_value(self, tzinfo: typing.Optional[datetime.timezone]) -> InstrumentValue:
        return InstrumentValue(
            value=self.level_eod,
            moment=datetime.datetime.combine(self.calc_date, datetime.time.min, tzinfo=tzinfo))


@dataclasses.dataclass
class IndexInfo(InstrumentInfoProvider):
    """ Container for index information.
    """
    msci_index_code: str
    index_name: str

    def __init__(self, *, msci_index_code: str, index_name: str):
        self.msci_index_code = str(msci_index_code)
        self.index_name = str(index_name)

    def __str__(self):
        return (f"MSCI index ("
                f"msci_index_code={self.msci_index_code}, "
                f"index_name={self.index_name})")

    @property
    def instrument_info(self) -> InstrumentInfo:
        return InstrumentInfo(code=self.msci_index_code, name=self.index_name)


class MsciIndexesInfoDownloadParameters(typing.NamedTuple):
    """ Container for ``MsciStringDataDownloader.download_instruments_info_string parameters``.
    """
    index_scope: Scopes
    index_market: Market
    index_size: Size
    index_style: Style
    index_suite: IndexSuite

    @classmethod
    def safe_create(
            cls: typing.Type['MsciIndexesInfoDownloadParameters'],
            *,
            index_scope: Scopes,
            index_market: Market,
            index_size: Size,
            index_style: Style,
            index_suite: IndexSuite) -> 'MsciIndexesInfoDownloadParameters':
        """ Create new instance of ``MsciIndexesInfoDownloadParameters`` with arguments check.

        :param index_scope: Index scope.
        :param index_market: Index market.
        :param index_size: Index size.
        :param index_style: Index style.
        :param index_suite: Index suite.
        :return: ``MsciIndexesInfoDownloadParameters`` instance.
        """
        if not isinstance(index_scope, Scopes):
            raise TypeError(f"'index_scope' is not Scopes: {index_scope!r}")
        if not isinstance(index_market, Market):
            raise TypeError(f"'index_market' is not Market: {index_market!r}")
        if not isinstance(index_size, Size):
            raise TypeError(f"'index_size' is not Size: {index_size!r}")
        if not isinstance(index_style, Style):
            raise TypeError(f"'index_style' is not Style: {index_style!r}")
        if not isinstance(index_suite, IndexSuite):
            raise TypeError(f"'index_suite' is not IndexSuite: {index_suite!r}")

        return cls(
            index_scope=index_scope,
            index_market=index_market,
            index_size=index_size,
            index_style=index_style,
            index_suite=index_suite)


@dataclasses.dataclass
class MsciIndexHistoryDownloadParameters(InstrumentHistoryDownloadParameters):
    """ Container for ``MsciStringDataDownloader.download_instrument_history_string parameters``.
    """
    index_code: Annotated[str, InstrumentInfoParameter(instrument_identity=True)]
    currency: Currency
    index_variant: IndexLevel

    def clone_with_instrument_info_parameters(
            self,
            info_download_parameters: typing.Optional[MsciIndexesInfoDownloadParameters],
            instrument_info: typing.Optional[IndexInfo]) -> 'MsciIndexHistoryDownloadParameters':
        return MsciIndexHistoryDownloadParameters.generate_from(self, info_download_parameters, instrument_info)

    # noinspection PyUnusedLocal
    @classmethod
    def generate_from(
            cls: typing.Type['MsciIndexHistoryDownloadParameters'],
            history_download_parameters: typing.Optional['MsciIndexHistoryDownloadParameters'],
            info_download_parameters: typing.Optional[MsciIndexesInfoDownloadParameters],
            instrument_info: typing.Optional[IndexInfo]) -> 'MsciIndexHistoryDownloadParameters':
        """ Create new history download parameters instance with data from its arguments.

        :param history_download_parameters: Optional instrument history download parameters for cloning.
        :param info_download_parameters: Optional instrument info download parameters for cloning.
        :param instrument_info: Optional instrument info for cloning.
        :return: Cloned history download parameters instance (self) with replacing some attributes from
            `info_download_parameters` and `instrument_info`.
        """
        return cls(
            index_code=((None if history_download_parameters is None else history_download_parameters.index_code)
                        if instrument_info is None
                        else instrument_info.msci_index_code),
            currency=(None if history_download_parameters is None else history_download_parameters.currency),
            index_variant=(None if history_download_parameters is None else history_download_parameters.index_variant)
        )

    @classmethod
    def safe_create(
            cls: typing.Type['MsciIndexHistoryDownloadParameters'],
            *,
            index_code: str,
            currency: Currency,
            index_variant: IndexLevel) -> 'MsciIndexHistoryDownloadParameters':
        """ Create new instance of ``MsciIndexHistoryDownloadParameters`` with arguments check.

        :param index_code: Index code.
        :param currency: Currency.
        :param index_variant: Index level.
        :return: ``MsciIndexHistoryDownloadParameters`` instance.
        """
        if not isinstance(currency, Currency):
            raise TypeError(f"'currency' is not Currency: {currency!r}")
        if not isinstance(index_variant, IndexLevel):
            raise TypeError(f"'index_variant' is not IndexLevel: {index_variant!r}")

        return cls(
            index_code=str(index_code),
            currency=currency,
            index_variant=index_variant)


class MsciDownloadParametersFactory(DownloadParametersFactory):
    """ Download parameters factories and generators for MSCI.
    """

    @property
    def download_history_parameters_class(self) -> typing.Type[MsciIndexHistoryDownloadParameters]:
        return MsciIndexHistoryDownloadParameters

    @property
    def download_history_parameters_factory(self) -> typing.Callable[..., MsciIndexHistoryDownloadParameters]:
        return MsciIndexHistoryDownloadParameters.safe_create

    @property
    def download_info_parameters_class(self):
        return MsciIndexesInfoDownloadParameters

    @property
    def download_info_parameters_factory(self) -> typing.Callable[..., typing.Any]:
        return MsciIndexesInfoDownloadParameters.safe_create

    def generate_history_download_parameters_from(
            self,
            history_download_parameters: typing.Optional[MsciIndexHistoryDownloadParameters],
            info_download_parameters: typing.Optional[MsciIndexesInfoDownloadParameters],
            instrument_info: typing.Optional[IndexInfo]) -> MsciIndexHistoryDownloadParameters:
        return MsciIndexHistoryDownloadParameters.generate_from(
            history_download_parameters,
            info_download_parameters,
            instrument_info)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import unittest

from sane_finances.annotations import Volatile


class TestVolatile(unittest.TestCase):

    def test_generate_Success(self):
        expected_result = 'OK'
        volatile = Volatile(generator=lambda _context: expected_result)

        result = volatile.generate({})

        self.assertEqual(expected_result, result)

    def test_generate_raiseWrongGenerator(self):

        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = Volatile(generator=None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Utility for download text from web.
"""

import typing
import logging
import urllib.parse
import urllib.request
import urllib.error
import http.client

from .downloader import Downloader, DownloadStringResult, DownloadError
from .cachers import BaseCacher

logging.getLogger().addHandler(logging.NullHandler())


class UrlDownloadStringResult(DownloadStringResult):
    """ Contains downloaded string
    and provides mechanism for feedback of downloaded string quality (actuality, correctness, fullness, etc.)
    to downloader (e.g. for cache and throttle management)
    """

    def __init__(
            self,
            downloaded_string: str,
            cacher: BaseCacher,
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str]):
        """ Initialize instance.

        :param downloaded_string: Downloaded string.
        :param cacher: Cacher used for storing downloaded strings.
        :param url: URL of request.
        :param parameters: Parameters of request.
        :param headers: Headers of request.
        """
        super().__init__(downloaded_string)

        self.cacher = cacher
        self.url = url
        self.parameters = parameters
        self.headers = headers

    def set_correctness(self, is_correct: bool):
        super().set_correctness(is_correct)

        if not self.is_correct:
            self.cacher.drop(self.url, self.parameters, self.headers)


class UrlDownloader(Downloader):
    """ Cacheable string (text) downloader from web using ``urllib``.
    """

    default_timeout_in_seconds: float = 600.0
    default_encoding: str = 'utf-8'

    def __init__(
            self,
            cacher: BaseCacher,
            timeout_in_seconds: float = default_timeout_in_seconds,
            encoding: str = default_encoding):
        """ Initialize downloader.

        :param cacher: Cacher used for storing downloaded strings.
        :param timeout_in_seconds: Timeout value for download request.
        :param encoding: Encoding of downloaded strings.
        """
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.cacher = cacher
        self.timeout_in_seconds = timeout_in_seconds
        self.encoding = encoding

        self._parameters: typing.List[typing.Tuple[str, str]] = []  # [(param_code, param_value)]
        self._headers: typing.Dict[str, str] = {}  # {header_code : header_value}

    @property
    def parameters(self) -> typing.List[typing.Tuple[str, str]]:
        return self._parameters

    @parameters.setter
    def parameters(self, value: typing.List[typing.Tuple[str, str]]) -> None:
        self._parameters = value

    @property
    def headers(self) -> typing.Dict[str, str]:
        return self._headers

    @headers.setter
    def headers(self, value: typing.Dict[str, str]) -> None:
        self._headers = value

    def read_string_from(self, request: urllib.request.Request, encoding: str) -> str:
        """ Read data from request as string

        Can be mocked in tests.

        :param request: ``urllib.request.Request`` instance.
        :param encoding: Encoding of downloaded string.
        :return: Downloaded string.
        """
        try:
            with urllib.request.urlopen(request, timeout=self.timeout_in_seconds) as response:
                data: bytearray = response.read()

        except (urllib.error.URLError, http.client.HTTPException) as ex:
            raise DownloadError() from ex

        self.logger.debug(f"Downloaded {len(data or '')} bytes")
        return data.decode(encoding)

    def build_query(self, url: str) -> urllib.request.Request:
        """ Build request object based on `parameters` and `headers` attributes.

        :param url: URL of request.
        :return: ``urllib.request.Request`` instance.
        """
        data = urllib.parse.urlencode(self.parameters)
        if data:
            url += '?' + data
        req = urllib.request.Request(url, headers=self.headers, method='GET')

        return req

    def download_string_impl(self, request: urllib.request.Request, encoding: str) -> str:
        """ Calls when real (not cached) download needed.

        :param request: ``urllib.request.Request`` instance.
        :param encoding: Encoding of downloaded string.
        :return: Downloaded string.
        """
        self.logger.info(f"Download from {request.full_url!r} to string")

        return self.read_string_from(request, encoding)

    def download_string(self, url: str, encoding: str = None) -> UrlDownloadStringResult:
        request = self.build_query(url)
        encoding = self.encoding if encoding is None else encoding

        parameters = self.parameters.copy()
        headers = self.headers.copy()

        got_from_cache, result = self.cacher.retrieve(
            url,
            parameters,
            headers,
            lambda _request=request, _encoding=encoding: self.download_string_impl(_request, _encoding))

        if got_from_cache:
            self.logger.info(f"Got string for {request.full_url!r} from cache")

        return UrlDownloadStringResult(result, self.cacher, url, parameters, headers)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Metadata for export currency rates from cbr.ru
"""

import decimal
import typing
import datetime
import enum
import dataclasses

from ...base import (
    InstrumentValue, InstrumentInfo, InstrumentValueProvider, InstrumentInfoProvider,
    InstrumentHistoryDownloadParameters, DownloadParametersFactory)
from ...inspection import InstrumentInfoParameter
from ....annotations import LEGACY_ANNOTATIONS

if LEGACY_ANNOTATIONS:  # pragma: no cover
    from ....annotations import Annotated
else:  # pragma: no cover
    from typing import Annotated


class RateFrequencies(enum.Enum):
    """ Frequencies of rate changing.
    """
    def __new__(cls, value: str, description: str):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = str(description)
        return obj

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}: " \
               f"'{self.value}' ('{self.description}')>"

    DAILY = ('0', 'Daily')
    MONTHLY = ('1', 'Monthly')


@dataclasses.dataclass
class CurrencyRateValue(InstrumentValueProvider):
    """ Container for currency rate history item.
    """
    date: datetime.date
    value: decimal.Decimal
    nominal: int
    currency_id: str

    def __init__(self,
                 *,
                 date: datetime.date,
                 value: decimal.Decimal,
                 nominal: int,
                 currency_id: str):
        if not isinstance(date, datetime.date):
            raise TypeError("value_date is not date")

        self.date = date
        self.value = decimal.Decimal(value)
        self.nominal = int(nominal)

        self.currency_id = str(currency_id)

    def get_instrument_value(self, tzinfo: typing.Optional[datetime.timezone]) -> InstrumentValue:
        return InstrumentValue(
            value=self.value/self.nominal,
            moment=datetime.datetime.combine(self.date, datetime.time.min, tzinfo=tzinfo))


@dataclasses.dataclass
class CurrencyInfo(InstrumentInfoProvider):
    """ Container for currency information.
    """
    currency_id: str
    name: str
    eng_name: str
    nominal: int
    parent_code: str

    def __init__(self, *, currency_id: str, name: str, eng_name: str, nominal: int, parent_code: str):
        self.currency_id = str(currency_id)
        self.name = str(name)
        self.eng_name = str(eng_name)
        self.nominal = int(nominal)
        self.parent_code = str(parent_code)

    def __str__(self):
        return (f"CBR currency ("
                f"currency_id={self.currency_id}, "
                f"name={self.name}, "
                f"eng_name={self.eng_name}, "
                f"nominal={self.nominal}, "
                f"parent_code={self.parent_code})")

    @property
    def instrument_info(self) -> InstrumentInfo:
        return InstrumentInfo(code=self.currency_id, name=self.name)


class CbrCurrenciesInfoDownloadParameters(typing.NamedTuple):
    """ Container for CbrStringDataDownloader.download_instruments_info_string parameters.
    """
    rate_frequency: RateFrequencies

    @classmethod
    def safe_create(
            cls: typing.Type['CbrCurrenciesInfoDownloadParameters'],
            *,
            rate_frequency: RateFrequencies) -> 'CbrCurrenciesInfoDownloadParameters':
        """ Create new instance of ``CbrCurrenciesInfoDownloadParameters`` with arguments check.

        :param rate_frequency: ``RateFrequencies`` value.
        :return: ``CbrCurrenciesInfoDownloadParameters`` instance.
        """
        if not isinstance(rate_frequency, RateFrequencies):
            raise TypeError(f"'rate_frequency' is not RateFrequencies: {rate_frequency!r}")

        # see https://github.com/PyCQA/pylint/issues/1801 for pylint disable hint details
        return cls(rate_frequency=RateFrequencies(rate_frequency))  # pylint: disable=no-value-for-parameter


@dataclasses.dataclass
class CbrCurrencyHistoryDownloadParameters(InstrumentHistoryDownloadParameters):
    """ Container for ``CbrStringDataDownloader.download_instrument_history_string`` parameters.
    """
    currency_id: Annotated[str, InstrumentInfoParameter(instrument_identity=True)]

    def clone_with_instrument_info_parameters(
            self,
            info_download_parameters: typing.Optional[CbrCurrenciesInfoDownloadParameters],
            instrument_info: typing.Optional[CurrencyInfo]) -> 'CbrCurrencyHistoryDownloadParameters':
        return CbrCurrencyHistoryDownloadParameters.generate_from(self, info_download_parameters, instrument_info)

    # noinspection PyUnusedLocal
    @classmethod
    def generate_from(
            cls: typing.Type['CbrCurrencyHistoryDownloadParameters'],
            history_download_parameters: typing.Optional['CbrCurrencyHistoryDownloadParameters'],
            info_download_parameters: typing.Optional[CbrCurrenciesInfoDownloadParameters],
            instrument_info: typing.Optional[CurrencyInfo]) -> 'CbrCurrencyHistoryDownloadParameters':
        """ Create new history download parameters instance with data from its arguments.

        :param history_download_parameters: Optional instrument history download parameters for cloning.
        :param info_download_parameters: Optional instrument info download parameters for cloning.
        :param instrument_info: Optional instrument info for cloning.
        :return: Cloned history download parameters instance (self) with replacing some attributes from
            `info_download_parameters` and `instrument_info`.
        """
        return cls(
            currency_id=((None if history_download_parameters is None else history_download_parameters.currency_id)
                         if instrument_info is None
                         else instrument_info.currency_id)
        )

    @classmethod
    def safe_create(
            cls: typing.Type['CbrCurrencyHistoryDownloadParameters'],
            *,
            currency_id: str) -> 'CbrCurrencyHistoryDownloadParameters':
        """ Create new instance of ``CbrCurrencyHistoryDownloadParameters`` with arguments check.

        :param currency_id: Currency ID value.
        :return: ``CbrCurrencyHistoryDownloadParameters`` instance.
        """
        return cls(currency_id=str(currency_id))


class CbrDownloadParametersFactory(DownloadParametersFactory):
    """ Download parameters factories and generators for Central Bank of Russia.
    """

    @property
    def download_history_parameters_class(self) -> typing.Type[CbrCurrencyHistoryDownloadParameters]:
        return CbrCurrencyHistoryDownloadParameters

    @property
    def download_history_parameters_factory(self) -> typing.Callable[..., CbrCurrencyHistoryDownloadParameters]:
        return CbrCurrencyHistoryDownloadParameters.safe_create

    @property
    def download_info_parameters_class(self):
        return CbrCurrenciesInfoDownloadParameters

    @property
    def download_info_parameters_factory(self) -> typing.Callable[..., typing.Any]:
        return CbrCurrenciesInfoDownloadParameters.safe_create

    def generate_history_download_parameters_from(
            self,
            history_download_parameters: typing.Optional[CbrCurrencyHistoryDownloadParameters],
            info_download_parameters: typing.Optional[CbrCurrenciesInfoDownloadParameters],
            instrument_info: typing.Optional[CurrencyInfo]) -> CbrCurrencyHistoryDownloadParameters:
        return CbrCurrencyHistoryDownloadParameters.generate_from(
            history_download_parameters,
            info_download_parameters,
            instrument_info)
